/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.example.learning;

import static fi.uef.envi.wavellite.entity.core.EntityFactory.component;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.componentPropertyTimePeriod;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.datasetObservation;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.situation;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.elementaryInfon;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.relation;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.relevantObject;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.relevantIndividual;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.dataset;

import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

import org.joda.time.DateTime;
import org.joda.time.format.DateTimeFormatter;
import org.joda.time.format.ISODateTimeFormat;

import com.espertech.esper.client.EventBean;
import com.espertech.esper.client.UpdateListener;

import fi.uef.envi.wavellite.entity.core.TemporalLocation;
import fi.uef.envi.wavellite.entity.core.base.TemporalLocationDateTime;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.situation.Attribute;
import fi.uef.envi.wavellite.entity.situation.Value;
import fi.uef.envi.wavellite.entity.situation.ElementaryInfon;
import fi.uef.envi.wavellite.entity.situation.Polarity;
import fi.uef.envi.wavellite.entity.situation.RelevantIndividual;
import fi.uef.envi.wavellite.entity.situation.Situation;
import fi.uef.envi.wavellite.module.learning.ModuleLearning;
import fi.uef.envi.wavellite.module.learning.base.AbstractModuleLearning;
import fi.uef.envi.wavellite.operator.extraction.base.esper.AbstractSituationExtractorListenerEsper;
import fi.uef.envi.wavellite.operator.extraction.base.esper.SituationExtractorEsper;

/**
 * <p>
 * Title: ExampleModuleLearningEsper
 * </p>
 * <p>
 * Description: In this example we use the Esper implementation of the knowledge
 * acquisition operator to acquire and represent situations from dataset
 * observations. Concretely, the example is for situations of 'strong winds'. We
 * have a 'strong wind' situation when wind speed data exceeds 10 m/s. A dataset
 * represents such data.
 * </p>
 * <p>
 * Project: Wavellite Example
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class ExampleModuleLearningEsper {

	private static final DateTimeFormatter dtf = ISODateTimeFormat.dateTime()
			.withOffsetParsed();

	public static void main(String[] args) {
		// Our dataset with wind speed data
		Dataset dataset = dataset("windSpeedData");

		// Populate the dataset with dataset observations for wind speed. Wind
		// speed is measured every hour on November 6, 2013 between 6 and 9 AM.
		Set<DatasetObservation> observations = new LinkedHashSet<DatasetObservation>();

		observations.add(datasetObservation(
				dataset,
				component(componentPropertyTimePeriod(),
						dtf.parseDateTime("2013-11-06T06:00:00.000+02:00")),
				component("windSpeed", 7.5)));

		observations.add(datasetObservation(
				dataset,
				component(componentPropertyTimePeriod(),
						dtf.parseDateTime("2013-11-06T07:00:00.000+02:00")),
				component("windSpeed", 6.5)));

		observations.add(datasetObservation(
				dataset,
				component(componentPropertyTimePeriod(),
						dtf.parseDateTime("2013-11-06T08:00:00.000+02:00")),
				component("windSpeed", 11.0)));

		observations.add(datasetObservation(
				dataset,
				component(componentPropertyTimePeriod(),
						dtf.parseDateTime("2013-11-06T09:00:00.000+02:00")),
				component("windSpeed", 9.3)));

		// Now use the learning module to acquire situations from the dataset
		ModuleLearning module = new ExampleModuleLearning();
		Queue<Situation> situations = module.result();

		// Hand over all dataset observations to the module
		module.considerAll(observations);

		// Print the situations
		print(situations);
	}

	// Prints the situation with infon strong winds at 11 m/s on November 6,
	// 2013 at 8 AM
	// s |= <<strongWinds, 11.0, 2013-11-06T08:00:00.000+02:00, 1>>
	private static void print(Queue<Situation> situations) {
		StringBuffer sb = new StringBuffer();

		for (Situation situation : situations) {
			sb.append("Situation" + "\r\n");
			for (ElementaryInfon infon : situation.getSupportedInfons()) {
				sb.append("  <<");
				sb.append(infon.getRelation().getId());
				sb.append(", ");

				// Individual for wind speed
				RelevantIndividual individual1 = (RelevantIndividual) infon
						.getRelevantObjects().get(0);
				// Individual for temporal location
				Attribute individual2 = (Attribute) infon.getRelevantObjects()
						.get(1);

				Double windSpeed = (Double) ((Value) ((Attribute) individual1
						.getAttributes().toArray()[0]).getValue()).getValue();
				DateTime time = ((TemporalLocationDateTime) individual2
						.getValue()).getValue();

				sb.append(windSpeed);
				sb.append(", ");
				sb.append(dtf.print(time));
				sb.append(", ");
				sb.append(infon.getPolarity().value());
				sb.append(">>" + "\r\n");
			}
		}

		System.out.println(sb);
	}

	private static class ExampleModuleLearning extends AbstractModuleLearning {

		private SituationExtractorEsper operator;

		public ExampleModuleLearning() {
			// We use a knowledge acquisition operator, specifically the Esper
			// implementation. Specify which listener should be called.
			operator = new SituationExtractorEsper();
			// Specify which listener should be called.
			operator.setListener(new AbstractSituationExtractorListenerEsper() {
				{
					setUpdateListener(new UpdateListener() {
						@Override
						public void update(EventBean[] newEvents,
								EventBean[] oldEvents) {
							@SuppressWarnings({ "unchecked", "rawtypes" })
							Map<String, Object> result = (Map) newEvents[0]
									.getUnderlying();

							situations.add(situation(elementaryInfon(
									relation("strongWinds"), Polarity.True,
									relevantIndividual((Double) result
											.get("windSpeed")),
									relevantObject((TemporalLocation) result
											.get("timePeriod")))));
						}
					});
				}

			});
			// Specify the time period component property of dataset
			// observations
			operator.setTimePeriodComponentProperty(componentPropertyTimePeriod());

			// Specify the EPL statement
			operator.addExpression("select components[0].componentPropertyValue.value as timePeriod, "
					+ "components[1].componentPropertyValue.value as windSpeed "
					+ "from DatasetObservationType "
					+ "where components[1].componentPropertyValue.value > 10.0");
		}

		@Override
		public void considerAll(Collection<DatasetObservation> observations) {
			Collection<Situation> situations = new HashSet<Situation>();

			operator.extract(observations, situations);

			queue.addAll(situations);
		}

	}

}
