/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.example.learning;

import static fi.uef.envi.wavellite.entity.core.EntityFactory.situation;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.elementaryInfon;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.relation;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.relevantIndividual;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.dataset;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.datasetObservation;
import static fi.uef.envi.wavellite.entity.core.EntityFactory.component;

import java.util.Collection;
import java.util.HashSet;
import java.util.Queue;
import java.util.Set;

import weka.classifiers.functions.MultilayerPerceptron;

import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.situation.Attribute;
import fi.uef.envi.wavellite.entity.situation.Value;
import fi.uef.envi.wavellite.entity.situation.ElementaryInfon;
import fi.uef.envi.wavellite.entity.situation.Polarity;
import fi.uef.envi.wavellite.entity.situation.RelevantIndividual;
import fi.uef.envi.wavellite.entity.situation.Situation;
import fi.uef.envi.wavellite.module.learning.ModuleLearning;
import fi.uef.envi.wavellite.module.learning.base.AbstractModuleLearning;
import fi.uef.envi.wavellite.operator.extraction.base.weka.AbstractSituationExtractorListenerWeka;
import fi.uef.envi.wavellite.operator.extraction.base.weka.SituationExtractorWeka;

/**
 * <p>
 * Title: ExampleModuleLearningWeka
 * </p>
 * <p>
 * Description: In this example we use the WEKA implementation of the knowledge
 * acquisition operator to acquire and represent situations from dataset
 * observations. Concretely, the example classifies dataset observations into
 * either class A or class B.
 * </p>
 * <p>
 * Project: Wavellite Example
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class ExampleModuleLearningWeka {

	public static void main(String[] args) {
		// Our dataset with dataset observations to be classified
		Dataset dataset = dataset("d1");

		// The dataset observations
		Set<DatasetObservation> observations = new HashSet<DatasetObservation>();

		observations.add(datasetObservation(dataset, component("p1", 6),
				component("p2", 12.0), component("p3", 18)));

		observations.add(datasetObservation(dataset, component("p1", -6),
				component("p2", -12.0), component("p3", -18)));

		// Now use the learning module to acquire situations from the dataset
		ModuleLearning module = new ExampleModuleLearning();
		Queue<Situation> situations = module.result();

		// Hand over all dataset observations to the module
		module.considerAll(observations);

		// Print the situations
		print(situations);
	}

	// Prints two situations with the
	private static void print(Queue<Situation> situations) {
		StringBuffer sb = new StringBuffer();

		for (Situation situation : situations) {
			sb.append("Situation" + "\r\n");
			for (ElementaryInfon infon : situation.getSupportedInfons()) {
				sb.append("  <<");
				sb.append(infon.getRelation().getId());
				sb.append(", ");

				// Individual for wind speed
				RelevantIndividual individual1 = (RelevantIndividual) infon
						.getRelevantObjects().get(0);

				String clsLabel = (String) ((Value) ((Attribute) individual1
						.getAttributes().toArray()[0]).getValue()).getValue();

				sb.append(clsLabel);
				sb.append(", ");
				sb.append(infon.getPolarity().value());
				sb.append(">>" + "\r\n");
			}
		}

		System.out.println(sb);
	}

	private static class ExampleModuleLearning extends AbstractModuleLearning {

		private SituationExtractorWeka operator;

		public ExampleModuleLearning() {
			// We use the information acquisition function, specifically the
			// WEKA implementation. We use a MLP classifier and provide the ARFF
			// data used to train the classifier
			try {
				operator = new SituationExtractorWeka(
						new MultilayerPerceptron(),
						"src/main/resources/example.arff");
			} catch (Exception e) {
				e.printStackTrace();
			}

			// Specify which listener should be called.
			operator.setListener(new AbstractSituationExtractorListenerWeka() {

				@Override
				public void onClassification(String label) {
					situations.add(situation(elementaryInfon(relation("class"),
							Polarity.True, relevantIndividual(label))));
				}
			});
		}

		@Override
		public void considerAll(Collection<DatasetObservation> observations) {
			Collection<Situation> situations = new HashSet<Situation>();
			
			operator.extract(observations, situations);
			
			queue.addAll(situations);
		}

	}

}
