/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.math.matrix;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.math3.linear.Array2DRowRealMatrix;
import org.apache.commons.math3.linear.RealMatrix;

import fi.uef.envi.wavellite.entity.derivation.ComponentProperty;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValue;
import fi.uef.envi.wavellite.entity.derivation.DataStructureDefinition;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueDouble;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueInteger;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetObservationBase;
import fi.uef.envi.wavellite.math.SortDirection;

/**
 * <p>
 * Title: Matrix
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Math
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class Matrix {

	private String datasetId;
	private Dataset dataset;
	private int dimensionality;
	private boolean isSorted = false;
	private boolean isInitialized = false;
	private ComponentProperty sortComponentProperty;
	private SortDirection sortDirection;
	private List<ComponentProperty> componentProperties;
	private Map<ComponentProperty, Integer> componentPropertyToOrderMap;
	private List<Map<String, String>> observationMetadata = new ArrayList<Map<String, String>>();
	private List<List<ComponentPropertyValue>> matrix = new ArrayList<List<ComponentPropertyValue>>();
	private List<List<Double>> realMatrix = new ArrayList<List<Double>>();
	private List<ComponentProperty> numericalComponentProperties = new ArrayList<ComponentProperty>();

	private static final Logger log = Logger.getLogger(Matrix.class.getName());

	public Matrix(Collection<DatasetObservation> observations) {
		this(observations.toArray(new DatasetObservation[] {}));
	}
	
	public Matrix(DatasetObservation... observations) {
		for (DatasetObservation observation : observations) {
			addDatasetObservation(observation);
		}
	}

	public Matrix(Dataset dataset) {
		this(dataset.getDataStructureDefinition());

		setDataset(dataset);
	}

	public Matrix(Dataset dataset, ComponentProperty... componentProperties) {
		this(dataset);

		setComponentProperties(componentProperties);
	}

	public Matrix(DataStructureDefinition structure) {
		if (structure != null)
			setComponentProperties(structure.getComponentPropertiesOrdered());
	}
	
	public boolean isSorted() {
		return isSorted;
	}
	
	public void setIsSorted(boolean isSorted) {
		this.isSorted = isSorted;
	}
	
	public void setSortComponentProperty(ComponentProperty property) {
		this.sortComponentProperty = property;
	}
	
	public ComponentProperty getSortComponentProperty() {
		return sortComponentProperty;
	}
	
	public void setSortDirection(SortDirection direction) {
		this.sortDirection = direction;
	}
	
	public SortDirection getSortDirection() {
		return sortDirection;
	}

	public void setDataset(Dataset dataset) {
		this.dataset = dataset;
		this.datasetId = dataset.getId();
	}

	public void setComponentProperties(ComponentProperty[] properties) {
		if (properties == null)
			return;

		setComponentProperties(Arrays.asList(properties));
	}

	public void setComponentProperties(Set<ComponentProperty> properties) {
		if (properties == null)
			return;

		setComponentProperties(properties.toArray(new ComponentProperty[] {}));
	}

	public void setComponentProperties(List<ComponentProperty> properties) {
		if (properties == null)
			return;

		this.componentProperties = properties;
		this.componentPropertyToOrderMap = new HashMap<ComponentProperty, Integer>();

		for (int i = 0; i < componentProperties.size(); i++) {
			componentPropertyToOrderMap.put(componentProperties.get(i), i);
		}

		this.dimensionality = componentProperties.size();
		this.isInitialized = true;
	}

	public void setNumericalComponentProperties(DatasetObservation observation) {
		for (ComponentProperty property : componentProperties) {
			ComponentPropertyValue value = observation
					.getComponentPropertyValue(property);

			if (isNumericalComponentPropertyValue(value))
				numericalComponentProperties.add(property);
		}
	}

	public void addDatasetObservation(DatasetObservation observation) {
		addDatasetObservation(matrix.size() + 1, observation);
	}

	public void addDatasetObservation(Double... values) {
		if (!isInitialized)
			throw new RuntimeException(
					"Cannot add dataset observation; set the component properties first [isInitialized = false]");
		if (dataset == null) {
			throw new RuntimeException(
					"Cannot add dataset observation [dataset = null]");
		}

		DatasetObservation observation = new DatasetObservationBase();
		observation.setDataset(dataset);

		if (values.length != componentProperties.size())
			throw new RuntimeException(
					"Cannot add dataset observation [values = " + values.length
							+ "; componentProperties = "
							+ componentProperties.size() + "]");

		for (int i = 0; i < componentProperties.size(); i++) {
			observation.addComponent(componentProperties.get(i),
					new ComponentPropertyValueDouble(values[i]));
		}

		addDatasetObservation(observation);
	}

	public void addDatasetObservation(int rowIndex,
			DatasetObservation observation) {
		Set<ComponentProperty> properties = observation
				.getComponentProperties();

		if (!isInitialized) {
			setComponentProperties(properties);
			setNumericalComponentProperties(observation);
		}

		dataset = observation.getDataset();

		if (datasetId == null) {
			if (dataset != null)
				datasetId = observation.getDataset().getId();
		}

		if (dimensionality != properties.size())
			throw new RuntimeException(
					"Cannot add observation to dataset [properties = "
							+ properties.size() + "; dimenstionality = "
							+ dimensionality + "]");

		if (!datasetId.equals(dataset.getId())) {
			throw new RuntimeException(
					"Dataset id of matrix does not match with dataset id of added observation ["
							+ datasetId + " != " + dataset.getId() + "]");
		}

		Map<String, String> map;

		map = new HashMap<String, String>();
		map.put("id", observation.getId());
		map.put("type", observation.getType());

		if (rowIndex > observationMetadata.size())
			observationMetadata.add(map);
		else
			observationMetadata.add(rowIndex, map);

		ComponentPropertyValue[] a = new ComponentPropertyValue[dimensionality];
		List<Double> b = new ArrayList<Double>();

		for (ComponentProperty componentProperty : properties) {
			Integer index = componentPropertyToOrderMap.get(componentProperty);

			if (index == null)
				throw new NullPointerException("[index = null]");

			ComponentPropertyValue componentPropertyValue = observation
					.getComponentPropertyValue(componentProperty);

			if (isNumericalComponentPropertyValue(componentPropertyValue)) {
				b.add(componentPropertyValue.getValueAsDouble());
			}

			a[index] = componentPropertyValue;
		}

		if (rowIndex > matrix.size())
			matrix.add(Arrays.asList(a));
		else
			matrix.add(rowIndex, Arrays.asList(a));

		realMatrix.add(b);
	}

	public void addDatasetObservations(Matrix matrix) {
		addDatasetObservations(matrix.iterator());
	}

	public void addDatasetObservations(Iterator<DatasetObservation> iterator) {
		while (iterator.hasNext()) {
			addDatasetObservation(iterator.next());
		}
	}

	public DatasetObservation getDatasetObservation(int rowIndex) {
		List<ComponentPropertyValue> values = matrix.get(rowIndex);

		String observationId = observationMetadata.get(rowIndex).get("id");
		String observationType = observationMetadata.get(rowIndex).get("type");

		DatasetObservation ret = new DatasetObservationBase(observationId,
				observationType);
		ret.setDataset(dataset);

		for (int i = 0; i < dimensionality; i++) {
			ComponentProperty property = componentProperties.get(i);
			ComponentPropertyValue value = values.get(i);

			ret.addComponent(property, value);
		}

		return ret;
	}

	public Matrix getDatasetObservations(int rowIndexFrom, int rowIndexTo) {
		Matrix ret = new Matrix();

		int rowIndex = rowIndexFrom;

		while (rowIndex < rowIndexTo) {
			ret.addDatasetObservation(getDatasetObservation(rowIndex++));
		}

		return ret;
	}

	public List<ComponentProperty> getComponentProperties() {
		return componentProperties;
	}

	public ComponentProperty getComponentProperty(int index) {
		return componentProperties.get(index);
	}

	public ComponentProperty getNumericalComponentProperty(int index) {
		return numericalComponentProperties.get(index);
	}

	public ComponentPropertyValue getComponentPropertyValue(int row,
			ComponentProperty property) {
		Integer column = componentPropertyToOrderMap.get(property);

		if (column == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Property not found in map [property = " + property
						+ "; componentPropertyToOrderMap = "
						+ componentPropertyToOrderMap + "]");

			return null;
		}

		return getComponentPropertyValue(row, column);
	}

	public ComponentPropertyValue getComponentPropertyValue(int row, int column) {
		int size = matrix.size();

		if (row >= size) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Row out of matrix size [row = " + row + "; size = "
						+ size + "]");

			return null;
		}

		return matrix.get(row).get(column);
	}

	public RealMatrix asRealMatrix() {
		double[][] a = new double[realMatrix.size()][numericalComponentProperties
				.size()];

		for (int i = 0; i < realMatrix.size(); i++) {
			a[i] = ArrayUtils.toPrimitive(realMatrix.get(i).toArray(
					new Double[] {}));
		}

		return new Array2DRowRealMatrix(a);
	}

	public int getDimensionality() {
		return dimensionality;
	}

	public Iterator<DatasetObservation> iterator() {
		if (matrix.size() == 0) {
			if (log.isLoggable(Level.INFO))
				log.info("Empty matrix; returned empty iterator");
			
			return Collections.emptyIterator();
		}
		
		if (dataset == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Dataset is null; returned empty iterator");
			
			return Collections.emptyIterator();
		}
		
		if (dataset == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Dataset is null; returned empty iterator");
			
			return Collections.emptyIterator();
		}
		
		return new MatrixIterator(matrix, dataset, componentProperties);
	}

	public Iterator<ComponentPropertyValue> columnIterator(int colIndex) {
		return new MatrixColumnIterator(matrix, colIndex);
	}

	public Iterator<ComponentPropertyValue> columnIterator(
			ComponentProperty property) {
		Integer colIndex = componentPropertyToOrderMap.get(property);

		if (colIndex == null)
			throw new NullPointerException("[colIndex = null]");

		return new MatrixColumnIterator(matrix, colIndex);
	}

	public Dataset getDataset() {
		return dataset;
	}

	public int size() {
		return matrix.size();
	}
	
	public boolean isEmpty() {
		return matrix.isEmpty();
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;

		result = prime * result + ((dataset == null) ? 0 : dataset.hashCode());
		result = prime * result + ((matrix == null) ? 0 : matrix.hashCode());

		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;

		Matrix other = (Matrix) obj;

		if (dataset == null) {
			if (other.dataset != null)
				return false;
		} else if (!dataset.equals(other.dataset))
			return false;

		if (matrix == null) {
			if (other.matrix != null)
				return false;
		} else if (!matrix.equals(other.matrix))
			return false;

		return true;
	}

	@Override
	public String toString() {
		return "DatasetObservationMatrix [dataset = " + dataset + "; matrix = "
				+ matrix + "]";
	}

	private boolean isNumericalComponentPropertyValue(
			ComponentPropertyValue value) {
		if (value.getClass().equals(ComponentPropertyValueDouble.class))
			return true;
		if (value.getClass().equals(ComponentPropertyValueInteger.class))
			return true;

		return false;
	}

}
