/*
 * Copyright (C) 2014 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.math.matrix.operation.base;

import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.math3.linear.RealMatrix;
import org.apache.commons.math3.stat.descriptive.UnivariateStatistic;

import fi.uef.envi.wavellite.entity.derivation.ComponentProperty;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValue;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueDouble;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetBase;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetObservationBase;
import fi.uef.envi.wavellite.math.matrix.Matrix;
import fi.uef.envi.wavellite.math.matrix.operation.OperationDescriptiveStatistic;

/**
 * <p>
 * Title: AbstractOperationDescriptiveStatistic
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Math
 * </p>
 * <p>
 * Copyright: Copyright (C) 2014
 * </p>
 * 
 * @author Markus Stocker
 */

public abstract class AbstractOperationDescriptiveStatistic implements
		OperationDescriptiveStatistic {

	protected UnivariateStatistic statistic;

	private static final Logger log = Logger
			.getLogger(AbstractOperationDescriptiveStatistic.class.getName());

	public DatasetObservation evaluate(Matrix matrix) {
		return evaluate(matrix, new DatasetBase(UUID.randomUUID().toString()));
	}
	
	public DatasetObservation evaluate(Matrix matrix,
			ComponentProperty temporalLocationProperty) {
		return evaluate(matrix, temporalLocationProperty, new DatasetBase(UUID
				.randomUUID().toString()));
	}
	
	public DatasetObservation evaluate(Matrix matrix, Dataset targetDataset) {
		DatasetObservation ret = new DatasetObservationBase();
		ret.setDataset(targetDataset);

		if (statistic == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("[statistic = null]");

			return ret;
		}

		RealMatrix realMatrix = matrix.asRealMatrix();

		for (int j = 0; j < realMatrix.getColumnDimension(); j++) {
			ret.addComponent(
					matrix.getNumericalComponentProperty(j),
					new ComponentPropertyValueDouble(statistic
							.evaluate(realMatrix.getColumn(j))));
		}

		return ret;
	}

	public DatasetObservation evaluate(Matrix matrix,
			ComponentProperty temporalLocationProperty, Dataset targetDataset) {
		if (matrix.size() == 0) {
			DatasetObservation ret = new DatasetObservationBase();
			ret.setDataset(targetDataset);

			if (log.isLoggable(Level.WARNING))
				log.warning("Trying to compute mean of empty matrix [ret = "
						+ ret + "]");

			return ret;
		}

		// First sort
		if (!matrix.isSorted()) {
			matrix = new OperationSort().temporalLocation(matrix,
					temporalLocationProperty, targetDataset);
		}

		// Retrieve the value of the temporal location of the first dataset
		// observation in the matrix
		ComponentPropertyValue temporalLocationValue = matrix
				.getDatasetObservation(0).getComponentPropertyValue(
						temporalLocationProperty);

		return evaluate(matrix, temporalLocationProperty,
				temporalLocationValue, targetDataset);
	}

	private DatasetObservation evaluate(Matrix matrix,
			ComponentProperty temporalLocationProperty,
			ComponentPropertyValue temporalLocationValue, Dataset targetDataset) {
		DatasetObservation ret = new DatasetObservationBase();
		ret.setDataset(targetDataset);

		ret.setComponentPropertyValue(temporalLocationProperty,
				temporalLocationValue);

		ret.addComponents(evaluate(matrix, targetDataset).getComponents());

		return ret;
	}

}
