/*
 * Copyright (C) 2014 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.math.matrix.operation.base;

import java.util.Set;

import org.apache.commons.collections4.IteratorUtils;

import fi.uef.envi.wavellite.entity.core.TemporalLocationVisitor;
import fi.uef.envi.wavellite.entity.core.base.TemporalLocationDateTime;
import fi.uef.envi.wavellite.entity.core.base.TemporalLocationVisitorBase;
import fi.uef.envi.wavellite.entity.derivation.ComponentProperty;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValue;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValueVisitor;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueDouble;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueInteger;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueTemporalLocation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueVisitorBase;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetObservationBase;
import fi.uef.envi.wavellite.math.matrix.Matrix;

/**
 * <p>
 * Title: OperationNumericalDifferentiation
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Math
 * </p>
 * <p>
 * Copyright: Copyright (C) 2014
 * </p>
 * 
 * @author Markus Stocker
 */

public class OperationNumericalDifferentiation {

	private Double valueX, valueY;
	private ComponentPropertyValueVisitor componentPropertyValueVisitor = new OperationComponentPropertyValueVisitor();
	private TemporalLocationVisitor temporalLocationVisitor = new OperationTemporalLocationVisitor();

	public Matrix evaluate(Matrix matrix, ComponentProperty componentProperty,
			Dataset targetDataset) {
		return evaluate(matrix, componentProperty, targetDataset, true);
	}

	/**
	 * Component property is the one used for the x-axis, mostly the temporal
	 * location. If computeDifferenceQuotient is false then only the function
	 * difference is calculated; otherwise the function difference is divided by
	 * the point difference.
	 * 
	 * @param matrix
	 * @param componentProperty
	 * @param computeDifferenceQuotient
	 * @return Matrix
	 */
	public Matrix evaluate(Matrix matrix, ComponentProperty componentProperty,
			Dataset targetDataset, boolean computeDifferenceQuotient) {
		Matrix ret = new Matrix();

		DatasetObservation[] observations = IteratorUtils.toArray(
				matrix.iterator(), DatasetObservation.class);

		for (int i = 0; i < observations.length - 1; i++) {
			// f'(x) ~= (f(x+h) - f(x)) / h
			DatasetObservation o1 = observations[i];
			DatasetObservation o2 = observations[i + 1];

			Set<ComponentProperty> propertiesO1 = o1.getComponentProperties();
			Set<ComponentProperty> propertiesO2 = o2.getComponentProperties();

			if (!propertiesO1.equals(propertiesO2)) {
				throw new RuntimeException("[" + propertiesO1 + " != "
						+ propertiesO2 + "]");
			}

			Set<ComponentProperty> properties = propertiesO1;

			ComponentPropertyValue componentPropertyValueX1 = o1
					.getComponentPropertyValue(componentProperty);
			componentPropertyValueX1.accept(componentPropertyValueVisitor);
			double valueX1 = valueX;
			ComponentPropertyValue componentPropertyValueX2 = o2
					.getComponentPropertyValue(componentProperty);
			componentPropertyValueX2.accept(componentPropertyValueVisitor);
			double valueX2 = valueX;

			double h = valueX2 - valueX1;

			DatasetObservation observation = new DatasetObservationBase();
			observation.setDataset(targetDataset);

			for (ComponentProperty property : properties) {
				if (property.equals(componentProperty)) {
					// This is the component property corresponding to the
					// x-axis; use the first value as its value
					observation.addComponent(componentProperty,
							componentPropertyValueX1);

					continue;
				}

				valueY = null;

				// Otherwise, get the values for this property for both o1 and
				// o2, if numerical to compute f(x+h) - f(x); if not numerical,
				// then add the corresponding value for o1 to the returned
				// observation (might be a spatial location or string or
				// something else).
				ComponentPropertyValue componentPropertyValueY1 = o1
						.getComponentPropertyValue(property);
				componentPropertyValueY1.accept(componentPropertyValueVisitor);

				if (valueY == null) {
					observation
							.addComponent(property, componentPropertyValueY1);

					continue;
				}

				double valueY1 = valueY;

				valueY = null;

				ComponentPropertyValue componentPropertyValueY2 = o2
						.getComponentPropertyValue(property);
				componentPropertyValueY2.accept(componentPropertyValueVisitor);

				if (valueY == null) {
					throw new RuntimeException("[value = null]");
				}

				double valueY2 = valueY;

				double f = valueY2 - valueY1;

				if (computeDifferenceQuotient)
					observation.addComponent(property,
							new ComponentPropertyValueDouble(f / h));
				else
					observation.addComponent(property,
							new ComponentPropertyValueDouble(f));
			}

			ret.addDatasetObservation(observation);
		}

		return ret;
	}

	private class OperationComponentPropertyValueVisitor extends
			ComponentPropertyValueVisitorBase {

		@Override
		public void visit(ComponentPropertyValueDouble componentPropertyValue) {
			valueY = componentPropertyValue.getValue();
		}

		@Override
		public void visit(ComponentPropertyValueInteger componentPropertyValue) {
			valueY = componentPropertyValue.getValueAsDouble();
		}

		@Override
		public void visit(
				ComponentPropertyValueTemporalLocation componentPropertyValue) {
			componentPropertyValue.getValueAsTemporalLocation().accept(
					temporalLocationVisitor);
		}
	}

	private class OperationTemporalLocationVisitor extends
			TemporalLocationVisitorBase {

		@Override
		public void visit(TemporalLocationDateTime time) {
			valueX = (double) time.getValueAsDateTime().getMillis();
		}
	}

}
