/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.math.matrix.operation.base;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.joda.time.DateTime;

import fi.uef.envi.wavellite.entity.derivation.ComponentProperty;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetBase;
import fi.uef.envi.wavellite.math.AggregationPeriod;
import fi.uef.envi.wavellite.math.matrix.Matrix;
import fi.uef.envi.wavellite.math.matrix.operation.MatrixOperation;

/**
 * <p>
 * Title: OperationSplit
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Math
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class OperationSplit implements MatrixOperation {

	private static final Logger log = Logger.getLogger(OperationSplit.class
			.getName());

	public Matrix[] evaluate(Matrix matrix,
			ComponentProperty temporalLocationProperty,
			AggregationPeriod aggregationPeriod) {
		return evaluate(matrix, temporalLocationProperty, aggregationPeriod,
				new DatasetBase(UUID.randomUUID().toString()));
	}

	/**
	 * Split a matrix according to a component property for temporal location
	 * 
	 * @param matrix
	 * @param componentProperty
	 * @param aggregationPeriod
	 * @param targetDataset
	 * @return Matrix[]
	 */
	public Matrix[] evaluate(Matrix matrix,
			ComponentProperty componentProperty,
			AggregationPeriod aggregationPeriod, Dataset targetDataset) {
		if (matrix.isEmpty())
			return new Matrix[] {};

		if (componentProperty == null)
			throw new NullPointerException("[componentProperty = null]");

		if (aggregationPeriod == null) {
			aggregationPeriod = AggregationPeriod.MINUTE;

			if (log.isLoggable(Level.WARNING))
				log.warning("Aggregation period is null; using "
						+ aggregationPeriod);
		}

		// Make sure the matrix is sorted
		if (!(matrix.isSorted() && matrix.getSortComponentProperty().equals(
				componentProperty))) {
			OperationSort operation = new OperationSort();
			matrix = operation.temporalLocation(matrix, componentProperty,
					targetDataset);
		}

		Iterator<DatasetObservation> it = matrix.iterator();

		List<Matrix> ret = new ArrayList<Matrix>();

		boolean isFirst = true;
		DateTime nextDateTime = new DateTime(1970, 1, 1, 0, 0, 0);
		DateTime currDateTime = null;
		Matrix subMatrix = new Matrix();

		while (it.hasNext()) {
			DatasetObservation observation = it.next();

			currDateTime = observation
					.getComponentPropertyValue(componentProperty)
					.getValueAsTemporalLocation().getValueAsDateTime();

			if (currDateTime.isEqual(nextDateTime)
					|| currDateTime.isAfter(nextDateTime)) {
				if (!isFirst) {
					ret.add(subMatrix);
					subMatrix = new Matrix();
				}

				if (aggregationPeriod.equals(AggregationPeriod.MINUTE)) {
					if (currDateTime.getSecondOfMinute() == 0)
						nextDateTime = currDateTime.plusMinutes(1);
					else
						nextDateTime = currDateTime.minuteOfHour()
								.roundCeilingCopy();
				} else if (aggregationPeriod.equals(AggregationPeriod.HOUR)) {
					if (currDateTime.getMinuteOfHour() == 0)
						nextDateTime = currDateTime.plusHours(1);
					else
						nextDateTime = currDateTime.hourOfDay()
								.roundCeilingCopy();
				} else if (aggregationPeriod.equals(AggregationPeriod.DAY)) {
					if (currDateTime.getHourOfDay() == 0)
						nextDateTime = currDateTime.plusDays(1);
					else
						nextDateTime = currDateTime.dayOfMonth()
								.roundCeilingCopy();
				} else {
					if (log.isLoggable(Level.WARNING))
						log.warning("Usupported aggregation period; using MINUTE as default [aggregationPeriod = "
								+ aggregationPeriod + "]");
					if (currDateTime.getSecondOfMinute() == 0)
						nextDateTime = currDateTime.plusMinutes(1);
					else
						nextDateTime = currDateTime.minuteOfHour()
								.roundCeilingCopy();
				}
			}

			observation.setDataset(targetDataset);
			subMatrix.addDatasetObservation(observation);

			isFirst = false;
		}

		ret.add(subMatrix);

		return ret.toArray(new Matrix[] {});
	}

}
