/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.operator.extraction.base.esper;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.joda.time.DateTime;

import com.espertech.esper.client.Configuration;
import com.espertech.esper.client.ConfigurationOperations;
import com.espertech.esper.client.EPAdministrator;
import com.espertech.esper.client.EPRuntime;
import com.espertech.esper.client.EPServiceProvider;
import com.espertech.esper.client.EPServiceProviderManager;
import com.espertech.esper.client.EPStatement;
import com.espertech.esper.client.UpdateListener;
import com.espertech.esper.client.time.CurrentTimeEvent;

import fi.uef.envi.wavellite.entity.core.TemporalLocation;
import fi.uef.envi.wavellite.entity.core.base.TemporalLocationDateTime;
import fi.uef.envi.wavellite.entity.derivation.Component;
import fi.uef.envi.wavellite.entity.derivation.ComponentProperty;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValue;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyBase;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueTemporalLocation;
import fi.uef.envi.wavellite.entity.situation.Situation;
import fi.uef.envi.wavellite.operator.extraction.SituationExtractor;
import fi.uef.envi.wavellite.vocabulary.SDMX;

/**
 * <p>
 * Title: SituationExtractorEsper
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Operator
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class SituationExtractorEsper implements
		SituationExtractor<SituationExtractorListenerEsper> {

	private EPServiceProvider serviceProvider;
	private EPAdministrator administrator;
	private EPRuntime runtime;
	private Configuration configuration;
	private Set<EPStatement> statements;
	private boolean internalTimerEnabled = false;
	private ComponentProperty timePeriodComponentProperty;
	private SituationExtractorListenerEsper listener;
	private UpdateListener updateListener;

	private static final Logger log = Logger
			.getLogger(SituationExtractorEsper.class.getName());

	public SituationExtractorEsper() {
		configuration = new Configuration();

		enableInternalTimer(internalTimerEnabled);

		serviceProvider = EPServiceProviderManager
				.getDefaultProvider(configuration);
		administrator = serviceProvider.getEPAdministrator();
		runtime = serviceProvider.getEPRuntime();
		statements = new HashSet<EPStatement>();

		initEventTypes();
		setTimePeriodComponentProperty(new ComponentPropertyBase(
				SDMX.Dimension.timePeriod));
	}

	public void setListener(SituationExtractorListenerEsper listener) {
		this.listener = listener;
		this.updateListener = listener.getUpdateListener();
	}

	public void addExpression(String expression) {
		addExpression(expression, null);
	}

	public void addExpression(String expression, UpdateListener listener) {
		EPStatement statement = administrator.createEPL(expression);

		if (listener != null) {
			statement.addListener(listener);
		} else if (updateListener != null) {
			statement.addListener(updateListener);
		} else {
			if (log.isLoggable(Level.WARNING))
				log.warning("No update listener [expression = " + expression
						+ "]");
		}

		statements.add(statement);
	}

	@Override
	public void extract(Collection<DatasetObservation> observations,
			Collection<Situation> situations) {
		if (listener == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("[listener = null]");
			return;
		}

		listener.setSituations(situations);

		for (DatasetObservation observation : observations) {
			setTime(observation);

			runtime.sendEvent(toEvent(observation), "DatasetObservationType");
		}
	}

	public void enableInternalTimer(boolean internalTimerEnabled) {
		configuration.getEngineDefaults().getThreading()
				.setInternalTimerEnabled(internalTimerEnabled);

		this.internalTimerEnabled = internalTimerEnabled;
	}

	public void setTimePeriodComponentProperty(ComponentProperty property) {
		this.timePeriodComponentProperty = property;
	}

	private void setTime(DatasetObservation observation) {
		if (internalTimerEnabled)
			return;

		ComponentPropertyValue value = observation
				.getComponentPropertyValue(timePeriodComponentProperty);

		if (value == null)
			throw new NullPointerException(
					"No time period component property [timePeriodComponentProperty = "
							+ timePeriodComponentProperty + "; observation = "
							+ observation + "]");

		if (!(value instanceof ComponentPropertyValueTemporalLocation))
			throw new RuntimeException("Expected temporal location [value = "
					+ value + "]");

		TemporalLocation location = ((ComponentPropertyValueTemporalLocation) value)
				.getValue();

		if (location == null)
			throw new NullPointerException("No value [value = " + value + "]");

		if (!(location instanceof TemporalLocationDateTime))
			throw new RuntimeException("Expected date time [location = "
					+ location + "]");

		DateTime dt = ((TemporalLocationDateTime) location).getValue();

		runtime.sendEvent(new CurrentTimeEvent(dt.getMillis()));
	}

	private static Map<String, Object> toEvent(DatasetObservation observation) {
		Map<String, Object> m1 = new HashMap<String, Object>();
		m1.put("id", observation.getId());
		m1.put("type", observation.getType());

		Map<String, Object> m2 = new HashMap<String, Object>();
		m1.put("dataset", m2);
		Dataset dataset = observation.getDataset();

		if (dataset == null) {
			if (log.isLoggable(Level.INFO))
				log.info("No dataset [observation = " + observation + "]");
		} else {
			m2.put("id", dataset.getId());
			m2.put("type", dataset.getType());
		}

		List<Map<String, Object>> m3 = new ArrayList<Map<String, Object>>();

		for (Component component : observation.getComponents()) {
			ComponentProperty property = component.getComponentProperty();
			ComponentPropertyValue value = component
					.getComponentPropertyValue();
			Map<String, Object> m4 = new HashMap<String, Object>();
			Map<String, Object> m5 = new HashMap<String, Object>();
			Map<String, Object> m6 = new HashMap<String, Object>();
			m3.add(m4);
			m4.put("componentProperty", m5);
			m4.put("componentPropertyValue", m6);
			m5.put("id", property.getId());
			m5.put("type", property.getType());
			m6.put("value", value.getValue());
		}

		m1.put("components", m3.toArray());

		return m1;
	}

	private void initEventTypes() {
		ConfigurationOperations configuration = administrator
				.getConfiguration();

		Map<String, Object> componentProperty = new HashMap<String, Object>();
		componentProperty.put("id", String.class);
		componentProperty.put("type", String.class);
		configuration.addEventType("ComponentPropertyType", componentProperty);

		Map<String, Object> componentPropertyValue = new HashMap<String, Object>();
		componentPropertyValue.put("value", Double.class);
		configuration.addEventType("ComponentPropertyValueType",
				componentPropertyValue);

		Map<String, Object> components = new HashMap<String, Object>();
		components.put("componentProperty", "ComponentPropertyType");
		components.put("componentPropertyValue", "ComponentPropertyValueType");
		configuration.addEventType("ComponentType", components);

		Map<String, Object> dataset = new HashMap<String, Object>();
		dataset.put("id", String.class);
		dataset.put("type", String.class);
		configuration.addEventType("DatasetType", components);

		Map<String, Object> datasetObservation = new HashMap<String, Object>();
		datasetObservation.put("id", String.class);
		datasetObservation.put("type", String.class);
		datasetObservation.put("dataset", "DatasetType");
		datasetObservation.put("components", "ComponentType[]");
		configuration
				.addEventType("DatasetObservationType", datasetObservation);
	}

}
