/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.operator.extraction.base.weka;

import java.util.Collection;
import java.util.logging.Level;
import java.util.logging.Logger;

import weka.classifiers.Classifier;
import weka.core.Instance;
import weka.core.Instances;
import weka.core.SparseInstance;
import weka.core.converters.ConverterUtils.DataSource;

import fi.uef.envi.wavellite.entity.derivation.Component;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValue;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValueVisitor;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueDouble;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueInteger;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueSpatialLocation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueString;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueTemporalLocation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueUri;
import fi.uef.envi.wavellite.entity.situation.Situation;
import fi.uef.envi.wavellite.operator.extraction.SituationExtractor;

/**
 * <p>
 * Title: SituationAcquirerWeka
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Operator
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class SituationExtractorWeka implements
		SituationExtractor<SituationExtractorListenerWeka> {

	private Classifier classifier;
	private Instances data;
	private int numAttributes;
	private Instance instance;
	private int index;
	private SituationExtractorListenerWeka listener;
	private ComponentPropertyValueVisitor componentPropertyValueVisitor;

	private static final Logger log = Logger
			.getLogger(SituationExtractorWeka.class.getName());

	public SituationExtractorWeka(Classifier classifier, String source)
			throws Exception {
		this(classifier, new DataSource(source));
	}

	public SituationExtractorWeka(Classifier classifier, DataSource source) {
		if (classifier == null)
			throw new NullPointerException("ModuleLearningWeka [classifier = "
					+ classifier + "]");
		if (source == null)
			throw new NullPointerException("ModuleLearningWeka [source = "
					+ source + "]");

		this.classifier = classifier;
		this.componentPropertyValueVisitor = new OperatorComponentPropertyValueVisitor();

		try {
			this.data = source.getDataSet();
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		// http://weka.wikispaces.com/Use+Weka+in+your+Java+code
		if (data.classIndex() == -1)
			data.setClassIndex(data.numAttributes() - 1);

		try {
			this.classifier.buildClassifier(data);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		numAttributes = data.numAttributes();
	}

	public void setListener(SituationExtractorListenerWeka listener) {
		this.listener = listener;
	}

	@Override
	public void extract(Collection<DatasetObservation> observations,
			Collection<Situation> situations) {
		if (listener == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("[listener = null]");

			return;
		}

		listener.setSituations(situations);

		for (DatasetObservation observation : observations) {
			String label = classify(observation);

			listener.onClassification(label);
		}
	}

	public String classify(DatasetObservation observation) {
		String clsLabel = null;

		int dimensionality = observation.getDimensionality();

		// -1 because the observation does not have a label, which the data does
		if (dimensionality != numAttributes - 1) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Dataset observation dimensionality error [dimensionality = "
						+ observation.getDimensionality()
						+ "; numAttributes = " + numAttributes + "]");

			return clsLabel;
		}

		instance = new SparseInstance(dimensionality);

		index = 0;

		for (Component component : observation.getComponents()) {
			ComponentPropertyValue value = component
					.getComponentPropertyValue();
			value.accept(componentPropertyValueVisitor);
		}

		try {
			double clsId = classifier.classifyInstance(instance);
			clsLabel = data.classAttribute().value((int) clsId);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}

		return clsLabel;
	}

	private class OperatorComponentPropertyValueVisitor implements
			ComponentPropertyValueVisitor {

		public void visit(ComponentPropertyValueDouble value) {
			instance.setValue(data.attribute(index++), value.getValue());

		}

		@Override
		public void visit(ComponentPropertyValueInteger value) {
			instance.setValue(data.attribute(index++), value.getValue());

		}

		@Override
		public void visit(ComponentPropertyValueTemporalLocation value) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Temporal location value unsupported in WEKA classification; classify dataset observations without components with temporal location value [value = "
						+ value + "]");
		}

		@Override
		public void visit(ComponentPropertyValueUri value) {
			if (log.isLoggable(Level.WARNING))
				log.warning("URI value unsupported in WEKA classification; classify dataset observations without components with URI value [value = "
						+ value + "]");
		}

		@Override
		public void visit(ComponentPropertyValueString value) {
			if (log.isLoggable(Level.WARNING))
				log.warning("String value unsupported in WEKA classification; classify dataset observations without components with string value [value = "
						+ value + "]");
		}

		@Override
		public void visit(ComponentPropertyValueSpatialLocation value) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Spatial location value unsupported in WEKA classification; classify dataset observations without components with spatial location value [value = "
						+ value + "]");
		}

	}

}
