/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.operator.translation.base;

import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;

import fi.uef.envi.wavellite.entity.core.EntityFactory;
import fi.uef.envi.wavellite.entity.core.TemporalLocation;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyDimension;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyMeasure;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueDouble;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueTemporalLocation;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetBase;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetObservationBase;
import fi.uef.envi.wavellite.entity.observation.ObservationValue;
import fi.uef.envi.wavellite.entity.observation.ObservationValueVisitor;
import fi.uef.envi.wavellite.entity.observation.SensorObservation;
import fi.uef.envi.wavellite.entity.observation.SensorOutput;
import fi.uef.envi.wavellite.entity.observation.base.ObservationValueDouble;
import fi.uef.envi.wavellite.operator.translation.SensorObservationTranslator;

/**
 * <p>
 * Title: SensorObservationTranslatorBase
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Operator
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class SensorObservationTranslatorBase implements
		SensorObservationTranslator {

	private Dataset targetDataset;
	private DatasetObservation ret;
	private ComponentPropertyDimension temporalLocationDimension;
	private ComponentPropertyMeasure sensorOutputMeasure;
	private String datasetObservationType;
	private ObservationValueVisitor observationValueVisitor;

	private final static Logger log = Logger
			.getLogger(SensorObservationTranslatorBase.class.getName());

	public SensorObservationTranslatorBase(
			ComponentPropertyMeasure sensorOutputMeasure, Dataset targetDataset) {
		this(EntityFactory.componentPropertyTimePeriod(), sensorOutputMeasure,
				targetDataset, null);
	}

	public SensorObservationTranslatorBase(
			ComponentPropertyDimension temporalLocationDimension,
			ComponentPropertyMeasure sensorOutputMeasure,
			Dataset targetDataset, String datasetObservationType) {
		this.temporalLocationDimension = temporalLocationDimension;
		this.sensorOutputMeasure = sensorOutputMeasure;
		this.targetDataset = targetDataset;
		this.datasetObservationType = datasetObservationType;
		this.observationValueVisitor = new TranslatorObservationValueVisitor();
	}

	@Override
	public DatasetObservation translate(SensorObservation observation) {
		ret = null;

		if (targetDataset == null) {
			targetDataset = new DatasetBase(UUID.randomUUID().toString());

			if (log.isLoggable(Level.WARNING))
				log.warning("Dataset is null. Create one with random UUID, which is probably not what you want [dataset = "
						+ targetDataset
						+ "; observation = "
						+ observation
						+ "]");
		}

		if (temporalLocationDimension == null)
			throw new NullPointerException(
					"[temporalLocationDimension = null; observation = "
							+ observation + "]");

		if (sensorOutputMeasure == null)
			throw new NullPointerException(
					"[sensorOutputMeasure = null; observation = " + observation
							+ "]");

		ret = new DatasetObservationBase();

		if (datasetObservationType != null)
			ret.setType(datasetObservationType);

		ret.setDataset(targetDataset);

		// The value of the dimension component property (necessarily time)
		TemporalLocation temporalLocation = observation.getTemporalLocation();

		ret.addComponent(temporalLocationDimension,
				new ComponentPropertyValueTemporalLocation(temporalLocation));

		// The value of the measure component property
		SensorOutput sensorOutput = observation.getSensorOutput();

		if (sensorOutput == null)
			throw new NullPointerException(
					"[sensorOutput = null; observation = " + observation + "]");

		ObservationValue observationValue = sensorOutput.getObservationValue();

		if (observationValue == null)
			throw new NullPointerException(
					"[observationValue = null; observation = " + observation
							+ "]");

		observationValue.accept(observationValueVisitor);

		return ret;
	}

	private class TranslatorObservationValueVisitor implements
			ObservationValueVisitor {

		@Override
		public void visit(ObservationValueDouble value) {
			ret.addComponent(sensorOutputMeasure,
					new ComponentPropertyValueDouble(value.getValue()));
		}

	}

	@Override
	public ComponentPropertyDimension getTemporalLocationDimension() {
		return temporalLocationDimension;
	}

	@Override
	public void setTemporalLocationDimension(ComponentPropertyDimension domain) {
		this.temporalLocationDimension = domain;
	}

	@Override
	public ComponentPropertyMeasure getSensorOutputMeasure() {
		return sensorOutputMeasure;
	}

	@Override
	public void setSensorOutputMeasure(ComponentPropertyMeasure range) {
		this.sensorOutputMeasure = range;
	}

	@Override
	public Dataset getTargetDataset() {
		return targetDataset;
	}

	@Override
	public void setTargetDataset(Dataset dataset) {
		this.targetDataset = dataset;
	}

	@Override
	public String getDatasetObservationType() {
		return datasetObservationType;
	}

	@Override
	public void setDatasetObservationType(String type) {
		this.datasetObservationType = type;
	}
}
