/*
 * Copyright (C) 2014 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.representation.csv;

//import java.util.HashMap;
//import java.util.Map;
import java.util.UUID;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.joda.time.format.DateTimeFormatter;
import org.joda.time.format.ISODateTimeFormat;

import fi.uef.envi.wavellite.entity.core.TemporalLocation;
import fi.uef.envi.wavellite.entity.core.base.TemporalLocationDateTime;
import fi.uef.envi.wavellite.entity.derivation.ComponentProperty;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValue;
import fi.uef.envi.wavellite.entity.derivation.DataStructureDefinition;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueDouble;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueInteger;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueTemporalLocation;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetBase;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetObservationBase;

/**
 * <p>
 * Title: EntityRepresentationCsvQb
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Representation
 * </p>
 * <p>
 * Copyright: Copyright (C) 2014
 * </p>
 * 
 * @author Markus Stocker
 */

public class EntityRepresentationCsvQb extends EntityRepresentationCsv {

	private Dataset dataset;
	// private DataStructureDefinition structure;
	private ComponentProperty[] componentProperties;
	private String columnSeparator;
	// private Map<Integer, Class<? extends ComponentPropertyValue>>
	// colToValueMap = new HashMap<Integer, Class<? extends
	// ComponentPropertyValue>>();
	private static final DateTimeFormatter dtf = ISODateTimeFormat.dateTime()
			.withOffsetParsed();

	private static final Logger log = Logger
			.getLogger(EntityRepresentationCsvQb.class.getName());

	public EntityRepresentationCsvQb() {

	}

	public EntityRepresentationCsvQb(Dataset dataset) {
		this(dataset, ",");
	}

	public EntityRepresentationCsvQb(Dataset dataset,
			ComponentProperty[] componentProperties) {
		this(dataset, componentProperties, ",");
	}

	public EntityRepresentationCsvQb(Dataset dataset,
			ComponentProperty[] componentProperties, String columnSeparator) {
		setDataset(dataset);
		setComponentProperties(componentProperties);
		setColumnSeparator(columnSeparator);
	}

	public EntityRepresentationCsvQb(Dataset dataset, String columnSeparator) {
		setDataset(dataset);
		setDataStructureDefinition(dataset.getDataStructureDefinition());
		setColumnSeparator(columnSeparator);
	}

	public void setDataset(Dataset dataset) {
		if (dataset == null) {
			dataset = new DatasetBase(UUID.randomUUID().toString());

			if (log.isLoggable(Level.WARNING))
				log.warning("Dataset is null; created one with random UUID [dataset = "
						+ dataset + "]");
		}

		this.dataset = dataset;
	}

	public void setDataStructureDefinition(DataStructureDefinition structure) {
		if (structure == null)
			throw new NullPointerException("[structure = " + structure + "]");

		setComponentProperties(structure.getComponentPropertiesOrdered()
				.toArray(new ComponentProperty[] {}));
	}

	public void setComponentProperties(ComponentProperty[] componentProperties) {
		if (componentProperties == null)
			throw new NullPointerException("[componentProperties = "
					+ componentProperties + "]");

		this.componentProperties = componentProperties;
	}

	public void setColumnSeparator(String columnSeparator) {
		if (columnSeparator == null) {
			columnSeparator = ",";

			if (log.isLoggable(Level.WARNING))
				log.warning("Column separator is null; using default [columnSeparator = "
						+ columnSeparator + "]");
		}

		this.columnSeparator = columnSeparator;
	}

	/**
	 * The parameter is expected to hold a list of component property values
	 * separated by a column separator, the default value of which is a comma.
	 * 
	 * @param string
	 * @return DatasetObservation
	 */
	@Override
	public DatasetObservation createDatasetObservation(String string) {
		DatasetObservation ret = new DatasetObservationBase();

		if (dataset == null)
			throw new NullPointerException("[dataset = null]");

		ret.setDataset(dataset);

		if (string == null)
			return ret;

		String[] columnValues = string.split(columnSeparator);

		if (columnValues.length != componentProperties.length)
			throw new RuntimeException("Component properties mismatch [cols = "
					+ columnValues.length + "; componentProperties = "
					+ componentProperties.length + "]");

		for (int columnIndex = 0; columnIndex < componentProperties.length; columnIndex++) {
			String columnValue = columnValues[columnIndex];
			ComponentProperty property = componentProperties[columnIndex];
			ComponentPropertyValue value = createComponentPropertyValue(columnValue);

			ret.addComponent(property, value);
		}

		return ret;
	}

	@Override
	public ComponentPropertyValue createComponentPropertyValue(String string) {
		Matcher matcher;

		// ISO 8601 date time (complete precision)
		matcher = Pattern
				.compile(
						"\\d{4}-[01]\\d-[0-3]\\dT[0-2]\\d:[0-5]\\d:[0-5]\\d\\.\\d+([+-][0-2]\\d:[0-5]\\d|Z)")
				.matcher(string);

		if (matcher.find()) {
			return createComponentPropertyValueTemporalLocation(matcher.group());
		}

		matcher = Pattern.compile("[-+]?[0-9]*\\.[0-9]+([eE][-+]?[0-9]+)?")
				.matcher(string);

		if (matcher.find()) {
			return createComponentPropertyValueDouble(matcher.group());
		}

		matcher = Pattern.compile("^[-+]?[0-9]*([eE][-+]?[0-9]+)?").matcher(
				string);

		if (matcher.find()) {
			return createComponentPropertyValueInteger(matcher.group());
		}

		throw new RuntimeException(
				"Unsupported component property value [string = " + string
						+ "]");
	}

	@Override
	public ComponentPropertyValueTemporalLocation createComponentPropertyValueTemporalLocation(
			String string) {
		return new ComponentPropertyValueTemporalLocation(
				createTemporalLocation(string));
	}

	@Override
	public TemporalLocation createTemporalLocation(String string) {
		Matcher matcher;

		matcher = Pattern
				.compile(
						"\\d{4}-[01]\\d-[0-3]\\dT[0-2]\\d:[0-5]\\d:[0-5]\\d\\.\\d+([+-][0-2]\\d:[0-5]\\d|Z)")
				.matcher(string);

		if (matcher.find())
			return createTemporalLocationDateTime(string);

		throw new RuntimeException("Unsupported temporal location [string = "
				+ string + "]");
	}

	@Override
	public TemporalLocationDateTime createTemporalLocationDateTime(String string) {
		return new TemporalLocationDateTime(dtf.parseDateTime(string));
	}

	@Override
	public ComponentPropertyValueDouble createComponentPropertyValueDouble(
			String string) {
		return new ComponentPropertyValueDouble(Double.valueOf(string));
	}

	@Override
	public ComponentPropertyValueInteger createComponentPropertyValueInteger(
			String string) {
		return new ComponentPropertyValueInteger(Integer.valueOf(string));
	}

}
