/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.representation.rdf;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.openrdf.model.Resource;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.vocabulary.OWL;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.RDFS;

import fi.uef.envi.wavellite.entity.core.SpatialGeometry;
import fi.uef.envi.wavellite.entity.core.SpatialLocation;
import fi.uef.envi.wavellite.entity.core.SpatialLocationVisitor;
import fi.uef.envi.wavellite.entity.core.base.SpatialGeometryLineString;
import fi.uef.envi.wavellite.entity.core.base.SpatialGeometryPoint;
import fi.uef.envi.wavellite.entity.core.base.SpatialGeometryPolygon;
import fi.uef.envi.wavellite.entity.core.base.SpatialLocationPlace;
import fi.uef.envi.wavellite.entity.core.base.SpatialLocationRegion;
import fi.uef.envi.wavellite.vocabulary.GeoSPARQL;
import fi.uef.envi.wavellite.vocabulary.SF;
import fi.uef.envi.wavellite.vocabulary.WOE;

/**
 * <p>
 * Title: EntityRepresentationRdfGeo
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Representation
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class EntityRepresentationRdfGeo extends EntityRepresentationRdf {

	private Set<Statement> statements;
	private SpatialLocationVisitor spatialLocationVisitor;
	private final static Logger log = Logger
			.getLogger(EntityRepresentationRdfGeo.class.getName());

	public EntityRepresentationRdfGeo() {
		this(null);
	}

	public EntityRepresentationRdfGeo(String ns) {
		super(ns);

		this.spatialLocationVisitor = new ThisSpatialLocationVisitor();
	}

	@Override
	public Set<Statement> createRepresentation(SpatialLocation location) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		location.accept(spatialLocationVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public SpatialLocation createSpatialLocation(Set<Statement> statements) {
		for (Statement statement : statements) {
			if (!statement.getPredicate().equals(RDF.TYPE))
				continue;

			Value o = statement.getObject();

			if (o.stringValue().equals(WOE.SpatialPlace))
				return createSpatialLocationPlace(statements);
			if (o.stringValue().equals(WOE.SpatialRegion))
				return createSpatialLocationRegion(statements);
		}

		if (log.isLoggable(Level.WARNING))
			log.warning("Could not create spatial location; expected types WOE.SpatialPlace or WOE.SpatialRegion [statements = "
					+ statements + "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(SpatialLocationPlace location) {
		Set<Statement> ret = new HashSet<Statement>();

		String locationId = location.getId();
		String label = location.getLabel();

		ret.add(_statement(_uri(locationId), RDF.TYPE, _uri(location.getType())));

		if (label == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("[label = null; location = " + location + "]");
		} else {
			ret.add(_statement(_uri(locationId), RDFS.LABEL, _literal(label)));

			for (URI sameAs : location.sameAs())
				ret.add(_statement(_uri(locationId), OWL.SAMEAS, sameAs));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public SpatialLocationPlace createSpatialLocationPlace(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			Value o = statement.getObject();

			if (!(statement.getPredicate().equals(RDF.TYPE) && o.stringValue()
					.equals(WOE.SpatialPlace)))
				continue;

			id = statement.getSubject().stringValue();
			type = o.stringValue();
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Cannot find spatial location place id; expected type WOE:SpatialLocation [statements = "
						+ statements + "]");

			return null;
		}

		// Add the null for the label, otherwise (String, String) is interpreted
		// as (id, label)
		SpatialLocationPlace ret = new SpatialLocationPlace(id, type, null);

		for (Statement statement : statements) {
			Resource s = statement.getSubject();
			URI p = statement.getPredicate();
			if (s.stringValue().equals(id) && p.equals(RDFS.LABEL))
				ret.setLabel(statement.getObject().stringValue());
			else if (s.stringValue().equals(id) && p.equals(OWL.SAMEAS))
				ret.addSameAs((URI) statement.getObject());
		}

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(SpatialLocationRegion location) {
		return createRepresentation(location, null);
	}

	public Set<Statement> createRepresentation(SpatialLocationRegion location,
			String literalType) {
		Set<Statement> ret = new HashSet<Statement>();

		String locationId = location.getId();

		ret.add(_statement(_uri(locationId), RDF.TYPE, _uri(location.getType())));
		// TODO: Should we materialize such inferences?
		ret.add(_statement(_uri(locationId), RDF.TYPE, WOE.asURI.SpatialRegion));

		SpatialGeometry spatialGeometry = location.getSpatialGeometry();

		if (spatialGeometry == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("[spatialGeometry = null; location = " + location
						+ "]");
		} else {
			ret.add(_statement(_uri(locationId), GeoSPARQL.asURI.hasGeometry,
					_uri(location.getSpatialGeometry().getId())));
			ret.addAll(createRepresentation(spatialGeometry, literalType));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public SpatialLocationRegion createSpatialLocationRegion(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			Value o = statement.getObject();

			if (!(statement.getPredicate().equals(RDF.TYPE) && o.stringValue()
					.equals(WOE.SpatialRegion)))
				continue;

			id = statement.getSubject().stringValue();
			type = o.stringValue();
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Cannot find spatial location region id; expected type WOE:SpatialRegion [statements = "
						+ statements + "]");

			return null;
		}

		SpatialLocationRegion ret = new SpatialLocationRegion(id, type);
		ret.setSpatialGeometry(createSpatialGeometry(statements));

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(SpatialGeometry spatialGeometry) {
		return createRepresentation(spatialGeometry, null);
	}

	public Set<Statement> createRepresentation(SpatialGeometry spatialGeometry,
			String literalType) {
		if (literalType == null)
			literalType = GeoSPARQL.wktLiteral;

		Set<Statement> ret = new HashSet<Statement>();

		String spatialGeometryId = spatialGeometry.getId();

		ret.add(_statement(_uri(spatialGeometryId), RDF.TYPE,
				_uri(spatialGeometry.getType())));

		String geometryAsString = null;

		if (literalType.equals(GeoSPARQL.wktLiteral))
			geometryAsString = spatialGeometry.getValueAsWKT();
		else if (literalType.equals(GeoSPARQL.gmlLiteral))
			geometryAsString = spatialGeometry.getValueAsGML();

		if (geometryAsString == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("[geometryAsString = null; spatialGeometry = "
						+ spatialGeometry + "]");
		} else {
			if (literalType.equals(GeoSPARQL.wktLiteral))
				ret.add(_statement(
						_uri(spatialGeometryId),
						GeoSPARQL.asURI.asWKT,
						_typedLiteral(geometryAsString,
								GeoSPARQL.asURI.wktLiteral)));
			else if (literalType.equals(GeoSPARQL.gmlLiteral))
				ret.add(_statement(
						_uri(spatialGeometryId),
						GeoSPARQL.asURI.asGML,
						_typedLiteral(geometryAsString,
								GeoSPARQL.asURI.gmlLiteral)));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public SpatialGeometry createSpatialGeometry(Set<Statement> statements) {
		for (Statement statement : statements) {
			URI p = statement.getPredicate();

			if (!p.equals(RDF.TYPE))
				continue;

			Value o = statement.getObject();

			if (o.stringValue().equals(SF.Point))
				return createSpatialGeometryPoint(statements);
			if (o.stringValue().equals(SF.LineString))
				return createSpatialGeometryLineString(statements);
			if (o.stringValue().equals(SF.Polygon))
				return createSpatialGeometryPolygon(statements);
		}

		if (log.isLoggable(Level.WARNING))
			log.warning("Cannot create spatial geometry; unrecognized type [statements = "
					+ statements + "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(
			SpatialGeometryPoint spatialGeometry) {
		return createRepresentation((SpatialGeometry) spatialGeometry);
	}

	@Override
	public SpatialGeometryPoint createSpatialGeometryPoint(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (!statement.getPredicate().equals(RDF.TYPE))
				continue;

			Value o = statement.getObject();

			if (o.stringValue().equals(SF.Point)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Cannot find spatial geometry point id [statements = "
						+ statements + "]");

			return null;
		}

		SpatialGeometryPoint ret = new SpatialGeometryPoint(id, type);

		for (Statement statement : statements) {
			if (!statement.getSubject().stringValue().equals(id))
				continue;

			URI p = statement.getPredicate();

			if (p.equals(GeoSPARQL.asURI.asWKT)) {
				ret.setValueAsWKT(statement.getObject().stringValue());
			} else if (p.equals(GeoSPARQL.asURI.asGML))
				ret.setValueAsGML(statement.getObject().stringValue());
		}

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(
			SpatialGeometryPolygon spatialGeometry) {
		return createRepresentation((SpatialGeometry) spatialGeometry);
	}

	@Override
	public SpatialGeometryPolygon createSpatialGeometryPolygon(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (!statement.getPredicate().equals(RDF.TYPE))
				continue;

			Value o = statement.getObject();

			if (o.stringValue().equals(SF.Polygon)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Cannot find spatial geometry polygon id [statements = "
						+ statements + "]");

			return null;
		}

		SpatialGeometryPolygon ret = new SpatialGeometryPolygon(id, type);

		for (Statement statement : statements) {
			if (!statement.getSubject().stringValue().equals(id))
				continue;

			URI p = statement.getPredicate();

			if (p.equals(GeoSPARQL.asURI.asWKT)) {
				ret.setValueAsWKT(statement.getObject().stringValue());
			} else if (p.equals(GeoSPARQL.asURI.asGML))
				ret.setValueAsGML(statement.getObject().stringValue());
		}

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(
			SpatialGeometryLineString spatialGeometry) {
		return createRepresentation((SpatialGeometry) spatialGeometry);
	}

	@Override
	public SpatialGeometryLineString createSpatialGeometryLineString(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (!statement.getPredicate().equals(RDF.TYPE))
				continue;

			Value o = statement.getObject();

			if (o.stringValue().equals(SF.LineString)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Cannot find spatial geometry line string id [statements = "
						+ statements + "]");

			return null;
		}

		SpatialGeometryLineString ret = new SpatialGeometryLineString(id, type);

		for (Statement statement : statements) {
			if (!statement.getSubject().stringValue().equals(id))
				continue;

			URI p = statement.getPredicate();

			if (p.equals(GeoSPARQL.asURI.asWKT)) {
				ret.setValueAsWKT(statement.getObject().stringValue());
			} else if (p.equals(GeoSPARQL.asURI.asGML))
				ret.setValueAsGML(statement.getObject().stringValue());
		}

		return ret;
	}

	private class ThisSpatialLocationVisitor implements SpatialLocationVisitor {

		@Override
		public void visit(SpatialLocationPlace location) {
			statements.addAll(createRepresentation(location));
		}

		@Override
		public void visit(SpatialLocationRegion location) {
			statements.addAll(createRepresentation(location));
		}

	}

}
