/*
 * Copyright (C) 2014 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.representation.rdf;

import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.vocabulary.RDF;

import fi.uef.envi.wavellite.entity.core.Entity;
import fi.uef.envi.wavellite.entity.core.Operation;
import fi.uef.envi.wavellite.entity.core.Operator;
import fi.uef.envi.wavellite.entity.core.Sensor;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.observation.Stimulus;
import fi.uef.envi.wavellite.entity.observation.SensorObservation;
import fi.uef.envi.wavellite.entity.situation.Attribute;
import fi.uef.envi.wavellite.entity.situation.AttributeVisitor;
import fi.uef.envi.wavellite.entity.situation.ElementaryInfon;
import fi.uef.envi.wavellite.entity.situation.Relation;
import fi.uef.envi.wavellite.entity.situation.RelevantIndividual;
import fi.uef.envi.wavellite.entity.situation.RelevantObject;
import fi.uef.envi.wavellite.entity.situation.RelevantObjectVisitor;
import fi.uef.envi.wavellite.entity.situation.Situation;
import fi.uef.envi.wavellite.entity.situation.Value;
import fi.uef.envi.wavellite.entity.situation.ValueVisitor;
import fi.uef.envi.wavellite.entity.situation.base.AttributeSpatialLocation;
import fi.uef.envi.wavellite.entity.situation.base.AttributeTemporalLocation;
import fi.uef.envi.wavellite.entity.situation.base.AttributeUri;
import fi.uef.envi.wavellite.entity.situation.base.AttributeValue;
import fi.uef.envi.wavellite.entity.situation.base.ValueDouble;
import fi.uef.envi.wavellite.entity.situation.base.ValueInteger;
import fi.uef.envi.wavellite.entity.situation.base.ValuePeriod;
import fi.uef.envi.wavellite.entity.situation.base.ValueString;
import fi.uef.envi.wavellite.entity.situation.base.ValueUri;
import fi.uef.envi.wavellite.vocabulary.PROV;

/**
 * <p>
 * Title: EntityRepresentationRdfProv
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Representation
 * </p>
 * <p>
 * Copyright: Copyright (C) 2014
 * </p>
 * 
 * @author Markus Stocker
 */

public class EntityRepresentationRdfProv extends EntityRepresentationRdf {

	private RelevantObjectVisitor relevantObjectVisitor;
	private AttributeVisitor attributeVisitor;
	private ValueVisitor attributeValueVisitor;
	private Set<Statement> statements;

	public EntityRepresentationRdfProv() {
		this(null);
	}

	public EntityRepresentationRdfProv(String ns) {
		super(ns);

		relevantObjectVisitor = new ThisRelevantObjectVisitor();
		attributeVisitor = new ThisAttributeVisitor();
		attributeValueVisitor = new ThisAttributeValueVisitor();
	}

	@Override
	public Set<Statement> createRepresentation(Sensor sensor) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(sensor.getId()), RDF.TYPE, PROV.asURI.Agent));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(Stimulus stimulus) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(stimulus.getId()), RDF.TYPE,
				PROV.asURI.Activity));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(SensorObservation observation) {
		Set<Statement> ret = new HashSet<Statement>();

		URI observationId = _uri(observation.getId());

		ret.add(_statement(observationId, RDF.TYPE, PROV.asURI.Entity));

		Sensor sensor = observation.getSensor();
		Stimulus stimulus = observation.getStimulus();

		if (sensor != null) {
			ret.add(_statement(observationId, PROV.asURI.wasAttributedTo,
					_uri(sensor.getId())));
			ret.addAll(createRepresentation(sensor));
		}

		if (stimulus != null) {
			ret.add(_statement(observationId, PROV.asURI.wasGeneratedBy,
					_uri(stimulus.getId())));
			ret.addAll(createRepresentation(stimulus));
		}

		if (sensor != null && stimulus != null) {
			ret.add(_statement(_uri(stimulus.getId()),
					PROV.asURI.wasAssociatedWith, _uri(sensor.getId())));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(Operation operation) {
		Set<Statement> ret = new HashSet<Statement>();

		URI operationId = _uri(operation.getId());

		ret.add(_statement(operationId, RDF.TYPE, PROV.asURI.Activity));

		Operator operator = operation.getAssociatedWith();

		if (operator != null) {
			ret.add(_statement(operationId, PROV.asURI.wasAssociatedWith,
					_uri(operator.getId())));
			ret.addAll(createRepresentation(operator));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(Operator operator) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(operator.getId()), RDF.TYPE, PROV.asURI.Agent));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(DatasetObservation observation) {
		return createRepresentation(_uri(observation.getId()),
				observation.getAttributedTo(), observation.getGeneratedBy(),
				observation.getDerivedFrom());
	}

	@Override
	public Set<Statement> createRepresentation(Situation situation) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.addAll(createRepresentation(_uri(situation.getId()),
				situation.getAttributedTo(), situation.getGeneratedBy(),
				situation.getDerivedFrom()));

		for (ElementaryInfon infon : situation.getSupportedInfons()) {
			ret.addAll(createRepresentation(infon));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(ElementaryInfon infon) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.addAll(createRepresentation(_uri(infon.getId()),
				infon.getAttributedTo(), infon.getGeneratedBy(),
				infon.getDerivedFrom()));

		Relation relation = infon.getRelation();

		if (relation != null)
			ret.addAll(createRepresentation(relation));

		for (RelevantObject object : infon.getRelevantObjects()) {
			ret.addAll(createRepresentation(object));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(Relation relation) {
		return createRepresentation(_uri(relation.getId()),
				relation.getAttributedTo(), relation.getGeneratedBy(),
				relation.getDerivedFrom());
	}

	@Override
	public Set<Statement> createRepresentation(RelevantObject object) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		object.accept(relevantObjectVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(RelevantIndividual individual) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.addAll(createRepresentation(_uri(individual.getId()),
				individual.getAttributedTo(), individual.getGeneratedBy(),
				individual.getDerivedFrom()));

		for (Attribute attribute : individual.getAttributes()) {
			ret.addAll(createRepresentation(attribute));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(Attribute attribute) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		attribute.accept(attributeVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(AttributeValue attribute) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.addAll(createRepresentation(_uri(attribute.getId()),
				attribute.getAttributedTo(), attribute.getGeneratedBy(),
				attribute.getDerivedFrom()));

		Value value = attribute.getValue();

		if (value != null) {
			ret.addAll(createRepresentation(value));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(
			AttributeSpatialLocation attribute) {
		return createRepresentation(_uri(attribute.getId()),
				attribute.getAttributedTo(), attribute.getGeneratedBy(),
				attribute.getDerivedFrom());
	}

	@Override
	public Set<Statement> createRepresentation(
			AttributeTemporalLocation attribute) {
		return createRepresentation(_uri(attribute.getId()),
				attribute.getAttributedTo(), attribute.getGeneratedBy(),
				attribute.getDerivedFrom());
	}

	@Override
	public Set<Statement> createRepresentation(AttributeUri attribute) {
		return createRepresentation(_uri(attribute.getId()),
				attribute.getAttributedTo(), attribute.getGeneratedBy(),
				attribute.getDerivedFrom());
	}

	@Override
	public Set<Statement> createRepresentation(Value value) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		value.accept(attributeValueVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(ValueDouble value) {
		return createRepresentation(_uri(value.getId()),
				value.getAttributedTo(), value.getGeneratedBy(),
				value.getDerivedFrom());
	}
	
	@Override
	public Set<Statement> createRepresentation(ValueString value) {
		return createRepresentation(_uri(value.getId()),
				value.getAttributedTo(), value.getGeneratedBy(),
				value.getDerivedFrom());
	}
	
	@Override
	public Set<Statement> createRepresentation(ValueUri value) {
		return createRepresentation(_uri(value.getId()),
				value.getAttributedTo(), value.getGeneratedBy(),
				value.getDerivedFrom());
	}
	
	@Override
	public Set<Statement> createRepresentation(ValueInteger value) {
		return createRepresentation(_uri(value.getId()),
				value.getAttributedTo(), value.getGeneratedBy(),
				value.getDerivedFrom());
	}

	@Override
	public Set<Statement> createRepresentation(ValuePeriod value) {
		return createRepresentation(_uri(value.getId()),
				value.getAttributedTo(), value.getGeneratedBy(),
				value.getDerivedFrom());
	}
	
	private Set<Statement> createRepresentation(URI entityId,
			Operator operator, Operation operation, List<Entity> derived) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(entityId, RDF.TYPE, PROV.asURI.Entity));

		if (operator != null) {
			ret.add(_statement(entityId, PROV.asURI.wasAttributedTo,
					_uri(operator.getId())));
			ret.addAll(createRepresentation(operator));
		}

		if (operation != null) {
			ret.add(_statement(entityId, PROV.asURI.wasGeneratedBy,
					_uri(operation.getId())));
			ret.addAll(createRepresentation(operation));
		}

		if (!(derived == null || derived.isEmpty())) {
			for (Entity entity : derived) {
				ret.add(_statement(entityId, PROV.asURI.wasDerivedFrom,
						_uri(entity.getId())));
				ret.addAll(createRepresentation(entity));
			}
		}

		return Collections.unmodifiableSet(ret);
	}

	private class ThisAttributeVisitor implements AttributeVisitor {

		@Override
		public void visit(AttributeValue attribute) {
			statements.addAll(createRepresentation(attribute));
		}

		@Override
		public void visit(AttributeSpatialLocation attribute) {
			statements.addAll(createRepresentation(attribute));
		}

		@Override
		public void visit(AttributeTemporalLocation attribute) {
			statements.addAll(createRepresentation(attribute));
		}

		@Override
		public void visit(AttributeUri attribute) {
			statements.addAll(createRepresentation(attribute));
		}

	}

	private class ThisAttributeValueVisitor implements ValueVisitor {

		@Override
		public void visit(ValueDouble value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ValueString value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ValueUri value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ValueInteger value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ValuePeriod value) {
			statements.addAll(createRepresentation(value));
		}

	}

	private class ThisRelevantObjectVisitor implements RelevantObjectVisitor {

		@Override
		public void visit(RelevantIndividual object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(AttributeSpatialLocation object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(AttributeTemporalLocation object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(AttributeUri object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValueDouble object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValueInteger object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValuePeriod object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValueString object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValueUri object) {
			statements.addAll(createRepresentation(object));
		}

	}
}
