/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.representation.rdf;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.openrdf.model.Literal;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.XMLSchema;

import fi.uef.envi.wavellite.entity.core.SpatialLocation;
import fi.uef.envi.wavellite.entity.core.TemporalLocation;
import fi.uef.envi.wavellite.entity.derivation.Component;
import fi.uef.envi.wavellite.entity.derivation.ComponentProperty;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValue;
import fi.uef.envi.wavellite.entity.derivation.ComponentPropertyValueVisitor;
import fi.uef.envi.wavellite.entity.derivation.ComponentSpecification;
import fi.uef.envi.wavellite.entity.derivation.DataStructureDefinition;
import fi.uef.envi.wavellite.entity.derivation.Dataset;
import fi.uef.envi.wavellite.entity.derivation.DatasetObservation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentBase;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyBase;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueDouble;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueInteger;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueSpatialLocation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueString;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueTemporalLocation;
import fi.uef.envi.wavellite.entity.derivation.base.ComponentPropertyValueUri;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetBase;
import fi.uef.envi.wavellite.entity.derivation.base.DatasetObservationBase;
import fi.uef.envi.wavellite.vocabulary.QB;
import fi.uef.envi.wavellite.vocabulary.WOE;

/**
 * <p>
 * Title: EntityRepresentationRdfQb
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Representation
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class EntityRepresentationRdfQb extends EntityRepresentationRdf {

	private String observationId;
	private String componentPropertyId;
	private Set<Statement> statements;
	private ComponentPropertyValueVisitor componentPropertyValueVisitor;
	private EntityRepresentationRdfTime entityRepresentationRdfTime;
	private EntityRepresentationRdfGeo entityRepresentationRdfGeo;

	private final static Logger log = Logger
			.getLogger(EntityRepresentationRdfQb.class.getName());

	public EntityRepresentationRdfQb() {
		this(null);
	}

	public EntityRepresentationRdfQb(String ns) {
		this(ns, null, null);
	}

	public EntityRepresentationRdfQb(String ns,
			EntityRepresentationRdfGeo entityRepresentationRdfGeo,
			EntityRepresentationRdfTime entityRepresentationRdfTime) {
		super(ns);

		if (entityRepresentationRdfGeo == null)
			this.entityRepresentationRdfGeo = new EntityRepresentationRdfGeo(ns);
		else
			this.entityRepresentationRdfGeo = entityRepresentationRdfGeo;

		if (entityRepresentationRdfGeo == null)
			this.entityRepresentationRdfTime = new EntityRepresentationRdfTime(
					ns);
		else
			this.entityRepresentationRdfTime = entityRepresentationRdfTime;

		this.componentPropertyValueVisitor = new ThisComponentPropertyValueVisitor();
	}

	@Override
	public void setNamespace(String ns) {
		super.setNamespace(ns);

		entityRepresentationRdfGeo.setNamespace(ns);
		entityRepresentationRdfTime.setNamespace(ns);
	}

	@Override
	public Set<Statement> createRepresentation(DatasetObservation observation) {
		Set<Statement> ret = new HashSet<Statement>();

		Dataset dataset = observation.getDataset();

		ret.add(_statement(_uri(observation.getId()), RDF.TYPE,
				_uri(observation.getType())));

		if (dataset == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No dataset [observation = " + observation + "]");
		} else {
			ret.add(_statement(_uri(observation.getId()), QB.asURI.dataSet,
					_uri(dataset.getId())));
			ret.addAll(createRepresentation(dataset));
		}

		observationId = observation.getId();

		for (Component component : observation.getComponents()) {
			ret.addAll(createRepresentation(component));
		}

		observationId = null;

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public DatasetObservation createDatasetObservation(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			Value o = statement.getObject();

			if (statement.getPredicate().equals(RDF.TYPE)
					&& o.stringValue().equals(WOE.DatasetObservation)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = o.stringValue();
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Statements contain more than one dataset observation [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot find dataset observation id; expected type WOE:DatasetObservation [statements = "
						+ statements + "]");

			return null;
		}

		DatasetObservation ret = new DatasetObservationBase(id, type);
		ret.setDataset(createDataset(statements));

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(QB.asURI.ComponentProperty)) {
				Set<Statement> result = new HashSet<Statement>();
				getTransitiveStatementsMatchingSubjectAndPredicate(statement
						.getSubject().stringValue(), statements, result);
				Component component = createComponent(result);

				if (component == null) {
					if (log.isLoggable(Level.WARNING))
						log.warning("Returned component is null [statements = "
								+ statements + "]");
				} else {
					ret.addComponent(component);
				}
			}
		}

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(Dataset dataset) {
		Set<Statement> ret = new HashSet<Statement>();

		String datasetId = dataset.getId();

		ret.add(_statement(_uri(datasetId), RDF.TYPE, QB.asURI.DataSet));

		DataStructureDefinition definition = dataset
				.getDataStructureDefinition();

		if (definition == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Dataset has no structure definition [dataset = "
						+ dataset + "]");
		} else {
			ret.add(_statement(_uri(datasetId), QB.asURI.structure,
					_uri(definition.getId())));
			ret.addAll(createRepresentation(definition));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Dataset createDataset(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			Value o = statement.getObject();

			if (statement.getPredicate().equals(RDF.TYPE)
					&& o.stringValue().equals(QB.DataSet)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = o.stringValue();
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Statements contain more than one dataset [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot find dataset id; expected type QB:DataSet [statements = "
						+ statements + "]");

			return null;
		}

		return new DatasetBase(id, type);
	}

	@Override
	public Set<Statement> createRepresentation(
			DataStructureDefinition definition) {
		Set<Statement> ret = new HashSet<Statement>();

		String definitionId = definition.getId();

		ret.add(_statement(_uri(definitionId), RDF.TYPE,
				_uri(definition.getType())));

		for (ComponentSpecification specification : definition
				.getComponentSpecifications()) {
			ret.add(_statement(_uri(definitionId), QB.asURI.component,
					_uri(specification.getId())));
			ret.addAll(createRepresentation(specification));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(
			ComponentSpecification specification) {
		Set<Statement> ret = new HashSet<Statement>();

		String specificationId = specification.getId();

		ret.add(_statement(_uri(specificationId), RDF.TYPE,
				_uri(specification.getType())));
		ret.add(_statement(
				_uri(specificationId),
				QB.asURI.order,
				_typedLiteral(Integer.toString(specification.getOrder()),
						XMLSchema.INT)));
		if (specification.isComponentRequired())
			ret.add(_statement(_uri(specificationId),
					QB.asURI.componentRequired,
					_typedLiteral(Boolean.TRUE.toString(), XMLSchema.BOOLEAN)));
		else
			ret.add(_statement(_uri(specificationId),
					QB.asURI.componentRequired,
					_typedLiteral(Boolean.FALSE.toString(), XMLSchema.BOOLEAN)));

		ComponentProperty property = specification.getComponentProperty();

		if (property == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Component specification has no component property [specification = "
						+ specification + "]");
		} else {
			String propertyId = property.getId();
			String propertyType = property.getType();

			if (propertyType.equals(QB.DimensionProperty))
				ret.add(_statement(_uri(specificationId), QB.asURI.dimension,
						_uri(propertyId)));
			else if (propertyType.equals(QB.MeasureProperty))
				ret.add(_statement(_uri(specificationId), QB.asURI.measure,
						_uri(propertyId)));
			else if (propertyType.equals(QB.AttributeProperty))
				ret.add(_statement(_uri(specificationId), QB.asURI.attribute,
						_uri(propertyId)));
			else
				ret.add(_statement(_uri(specificationId),
						QB.asURI.componentProperty, _uri(property.getId())));

			ret.addAll(createRepresentation(property));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(Component component) {
		Set<Statement> ret = new HashSet<Statement>();

		ComponentProperty property = component.getComponentProperty();

		if (property == null)
			return ret;

		componentPropertyId = property.getId();

		ret.addAll(createRepresentation(property));
		ret.addAll(createRepresentation(component.getComponentPropertyValue()));

		componentPropertyId = null;

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Component createComponent(Set<Statement> statements) {
		return new ComponentBase(createComponentProperty(statements),
				createComponentPropertyValue(statements));
	}

	@Override
	public Set<Statement> createRepresentation(ComponentProperty property) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(property.getId()), RDF.TYPE,
				_uri(property.getType())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ComponentProperty createComponentProperty(Set<Statement> statements) {
		componentPropertyId = null;
		String type = null;

		for (Statement statement : statements) {
			Value o = statement.getObject();
			if (statement.getPredicate().equals(RDF.TYPE)
					&& o.equals(QB.asURI.ComponentProperty)) {
				if (componentPropertyId == null) {
					componentPropertyId = statement.getSubject().stringValue();
					type = o.stringValue();
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Multiple component properties found in statements [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (componentPropertyId == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Cannot find component property id; expected type QB.ComponentProperty [statements = "
						+ statements + "]");
			return null;
		}

		return new ComponentPropertyBase(componentPropertyId, type);
	}

	@Override
	public Set<Statement> createRepresentation(ComponentPropertyValue value) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		value.accept(componentPropertyValueVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ComponentPropertyValue createComponentPropertyValue(
			Set<Statement> statements) {
		if (componentPropertyId == null)
			throw new NullPointerException("[componentPropertyId = null]");

		for (Statement statement : statements) {
			if (!statement.getPredicate().stringValue()
					.equals(componentPropertyId))
				continue;

			Value o = statement.getObject();

			if (o instanceof Literal) {
				URI dt = ((Literal) o).getDatatype();
				if (dt.equals(XMLSchema.DOUBLE))
					return createComponentPropertyValueDouble(Collections
							.singleton(statement));
				if (dt.equals(XMLSchema.INT) || dt.equals(XMLSchema.INTEGER))
					return createComponentPropertyValueInteger(Collections
							.singleton(statement));
				if (dt.equals(XMLSchema.STRING))
					return createComponentPropertyValueString(Collections
							.singleton(statement));

				if (log.isLoggable(Level.SEVERE))
					log.severe("Unrecognized component property value datatype [statements = "
							+ statements + "]");

				return null;
			}

			// Handle URI, TemporalLocation, and SpatialLocation
			if (o instanceof URI) {
				// Test if the uri is a temporal location
				if (statements.contains(f.createStatement((URI) o, RDF.TYPE,
						WOE.asURI.TimePoint))
						|| statements.contains(f.createStatement((URI) o,
								RDF.TYPE, WOE.asURI.TimeInterval)))
					return createComponentPropertyValueTemporalLocation(statements);
				if (statements.contains(f.createStatement((URI) o, RDF.TYPE,
						WOE.asURI.SpatialPlace))
						|| statements.contains(f.createStatement((URI) o,
								RDF.TYPE, WOE.asURI.SpatialRegion)))
					return createComponentPropertyValueSpatialLocation(statements);

				return createComponentPropertyValueUri(Collections
						.singleton(statement));
			}
		}

		if (log.isLoggable(Level.SEVERE))
			log.severe("Could not create component property value [statement = "
					+ statements + "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(
			ComponentPropertyValueDouble value) {
		if (observationId == null)
			throw new NullPointerException("[observationId = null]");
		if (componentPropertyId == null)
			throw new NullPointerException("[componentPropertyId = null]");

		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(observationId), _uri(componentPropertyId),
				_typedLiteral(value.getValue().toString(), XMLSchema.DOUBLE)));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ComponentPropertyValueDouble createComponentPropertyValueDouble(
			Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Expected singleton statement [statements = "
						+ statements + "]");

			return null;
		}

		return new ComponentPropertyValueDouble(Double.valueOf(statements
				.toArray(new Statement[] {})[0].getObject().stringValue()));
	}

	@Override
	public Set<Statement> createRepresentation(
			ComponentPropertyValueInteger value) {
		if (observationId == null)
			throw new NullPointerException("[observationId = null]");
		if (componentPropertyId == null)
			throw new NullPointerException("[componentPropertyId = null]");

		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(observationId), _uri(componentPropertyId),
				_typedLiteral(value.getValue().toString(), XMLSchema.INT)));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ComponentPropertyValueInteger createComponentPropertyValueInteger(
			Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Expected singleton statement [statements = "
						+ statements + "]");

			return null;
		}

		return new ComponentPropertyValueInteger(Integer.valueOf(statements
				.toArray(new Statement[] {})[0].getObject().stringValue()));
	}

	@Override
	public Set<Statement> createRepresentation(
			ComponentPropertyValueString value) {
		if (observationId == null)
			throw new NullPointerException("[observationId = null]");
		if (componentPropertyId == null)
			throw new NullPointerException("[componentPropertyId = null]");

		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(observationId), _uri(componentPropertyId),
				_typedLiteral(value.getValue().toString(), XMLSchema.STRING)));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ComponentPropertyValueString createComponentPropertyValueString(
			Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Expected singleton statement [statements = "
						+ statements + "]");

			return null;
		}

		return new ComponentPropertyValueString(
				statements.toArray(new Statement[] {})[0].getObject()
						.stringValue());
	}

	@Override
	public Set<Statement> createRepresentation(ComponentPropertyValueUri value) {
		if (observationId == null)
			throw new NullPointerException("[observationId = null]");
		if (componentPropertyId == null)
			throw new NullPointerException("[componentPropertyId = null]");

		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(observationId), _uri(componentPropertyId),
				value.getValue()));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ComponentPropertyValueUri createComponentPropertyValueUri(
			Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Expected singleton statement [statements = "
						+ statements + "]");

			return null;
		}

		return new ComponentPropertyValueUri(f.createURI(statements
				.toArray(new Statement[] {})[0].getObject().stringValue()));
	}

	@Override
	public Set<Statement> createRepresentation(
			ComponentPropertyValueTemporalLocation value) {
		return createRepresentation(value.getValue());
	}

	@Override
	public ComponentPropertyValueTemporalLocation createComponentPropertyValueTemporalLocation(
			Set<Statement> statements) {
		return new ComponentPropertyValueTemporalLocation(
				entityRepresentationRdfTime.createTemporalLocation(statements));
	}

	@Override
	public Set<Statement> createRepresentation(TemporalLocation location) {
		if (observationId == null)
			throw new NullPointerException("[observationId = null]");
		if (componentPropertyId == null)
			throw new NullPointerException("[componentPropertyId = null]");

		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(observationId), _uri(componentPropertyId),
				_uri(location.getId())));

		ret.addAll(entityRepresentationRdfTime.createRepresentation(location));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(
			ComponentPropertyValueSpatialLocation value) {
		return createRepresentation(value.getValue());
	}

	@Override
	public ComponentPropertyValueSpatialLocation createComponentPropertyValueSpatialLocation(
			Set<Statement> statements) {
		return new ComponentPropertyValueSpatialLocation(
				entityRepresentationRdfGeo.createSpatialLocation(statements));
	}

	@Override
	public Set<Statement> createRepresentation(SpatialLocation location) {
		if (observationId == null)
			throw new NullPointerException("[observationId = null]");
		if (componentPropertyId == null)
			throw new NullPointerException("[componentPropertyId = null]");

		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(observationId), _uri(componentPropertyId),
				_uri(location.getId())));

		ret.addAll(entityRepresentationRdfGeo.createRepresentation(location));

		return Collections.unmodifiableSet(ret);
	}

	private class ThisComponentPropertyValueVisitor implements
			ComponentPropertyValueVisitor {

		@Override
		public void visit(ComponentPropertyValueDouble value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ComponentPropertyValueInteger value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ComponentPropertyValueTemporalLocation value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ComponentPropertyValueSpatialLocation value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ComponentPropertyValueUri value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ComponentPropertyValueString value) {
			statements.addAll(createRepresentation(value));
		}

	}

}
