/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.representation.rdf;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.openrdf.model.Literal;
import org.openrdf.model.Resource;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.XMLSchema;

import fi.uef.envi.wavellite.entity.core.Feature;
import fi.uef.envi.wavellite.entity.core.Property;
import fi.uef.envi.wavellite.entity.core.Sensor;
import fi.uef.envi.wavellite.entity.core.SpatialLocation;
import fi.uef.envi.wavellite.entity.core.SpatialLocationVisitor;
import fi.uef.envi.wavellite.entity.core.TemporalLocation;
import fi.uef.envi.wavellite.entity.core.base.FeatureBase;
import fi.uef.envi.wavellite.entity.core.base.PropertyBase;
import fi.uef.envi.wavellite.entity.core.base.SensorBase;
import fi.uef.envi.wavellite.entity.core.base.SpatialLocationPlace;
import fi.uef.envi.wavellite.entity.core.base.SpatialLocationRegion;
import fi.uef.envi.wavellite.entity.observation.ObservationValue;
import fi.uef.envi.wavellite.entity.observation.ObservationValueVisitor;
import fi.uef.envi.wavellite.entity.observation.Stimulus;
import fi.uef.envi.wavellite.entity.observation.SensorObservation;
import fi.uef.envi.wavellite.entity.observation.SensorOutput;
import fi.uef.envi.wavellite.entity.observation.base.ObservationValueDouble;
import fi.uef.envi.wavellite.entity.observation.base.SensorObservationBase;
import fi.uef.envi.wavellite.entity.observation.base.SensorOutputBase;
import fi.uef.envi.wavellite.vocabulary.DUL;
import fi.uef.envi.wavellite.vocabulary.SSN;
import fi.uef.envi.wavellite.vocabulary.WOE;

/**
 * <p>
 * Title: EntityRepresentationRdfSsn
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Representation
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class EntityRepresentationRdfSsn extends EntityRepresentationRdf {

	private final static Logger log = Logger
			.getLogger(EntityRepresentationRdfSsn.class.getName());

	private ObservationValueVisitor observationValueVisitor;
	private ThisSpatialLocationVisitor spatialLocationVisitor;
	private Set<Statement> statements;
	private String observationId;
	private EntityRepresentationRdfGeo entityRepresentationRdfGeo;
	private EntityRepresentationRdfTime entityRepresentationRdfTime;

	public EntityRepresentationRdfSsn() {
		this(null);
	}

	public EntityRepresentationRdfSsn(String ns) {
		this(ns, null, null);
	}

	public EntityRepresentationRdfSsn(String ns,
			EntityRepresentationRdfGeo entityRepresentationRdfGeo,
			EntityRepresentationRdfTime entityRepresentationRdfTime) {
		super(ns);

		if (entityRepresentationRdfGeo == null)
			this.entityRepresentationRdfGeo = new EntityRepresentationRdfGeo(ns);
		else
			this.entityRepresentationRdfGeo = entityRepresentationRdfGeo;

		if (entityRepresentationRdfTime == null)
			this.entityRepresentationRdfTime = new EntityRepresentationRdfTime(
					ns);
		else
			this.entityRepresentationRdfTime = entityRepresentationRdfTime;

		this.spatialLocationVisitor = new ThisSpatialLocationVisitor();
		this.observationValueVisitor = new ThisObservationValueVisitor();
	}

	@Override
	public void setNamespace(String ns) {
		super.setNamespace(ns);

		entityRepresentationRdfGeo.setNamespace(ns);
		entityRepresentationRdfTime.setNamespace(ns);
	}

	@Override
	public Set<Statement> createRepresentation(Sensor sensor) {
		Set<Statement> ret = new HashSet<Statement>();

		URI sensorId = _uri(sensor.getId());

		ret.add(_statement(sensorId, RDF.TYPE, _uri(sensor.getType())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Sensor createSensor(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			URI p = statement.getPredicate();
			Value o = statement.getObject();

			if (p.equals(RDF.TYPE) && o.equals(SSN.asURI.Sensor)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find sensor id; expected type SSN:Sensor [statements = "
						+ statements + "]");

			return null;
		}

		return new SensorBase(id, type);
	}

	@Override
	public Set<Statement> createRepresentation(Stimulus stimulus) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(stimulus.getId()), RDF.TYPE,
				_uri(stimulus.getType())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(Property property) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(property.getId()), RDF.TYPE,
				_uri(property.getType())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Property createProperty(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			URI p = statement.getPredicate();
			Value o = statement.getObject();

			if (p.equals(RDF.TYPE) && o.equals(SSN.asURI.Property)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find property id; expected type SSN:Property [statements = "
						+ statements + "]");

			return null;
		}

		return new PropertyBase(id, type);
	}

	@Override
	public Set<Statement> createRepresentation(Feature feature) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(feature.getId()), RDF.TYPE,
				_uri(feature.getType())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Feature createFeature(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			URI p = statement.getPredicate();
			Value o = statement.getObject();

			if (p.equals(RDF.TYPE) && o.equals(SSN.asURI.FeatureOfInterest)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find feature id; expected type SSN:FeatureOfInterest [statements = "
						+ statements + "]");

			return null;
		}

		return new FeatureBase(id, type);
	}

	@Override
	public Set<Statement> createRepresentation(SensorOutput output) {
		Set<Statement> ret = new HashSet<Statement>();

		ObservationValue value = output.getObservationValue();

		ret.add(_statement(_uri(output.getId()), RDF.TYPE,
				_uri(output.getType())));

		if (value == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No observation value [output = " + output + "]");
		} else {
			ret.add(_statement(_uri(output.getId()), DUL.asURI.hasRegion,
					_uri(value.getId())));
			ret.addAll(createRepresentation(value));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public SensorOutput createSensorOutput(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			URI p = statement.getPredicate();
			Value o = statement.getObject();

			if (p.equals(RDF.TYPE) && o.equals(SSN.asURI.SensorOutput)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find sensor output id; expected type SSN:SensorOutput [statements = "
						+ statements + "]");

			return null;
		}

		return new SensorOutputBase(id, type,
				createObservationValue(statements));
	}

	@Override
	public Set<Statement> createRepresentation(ObservationValue value) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		value.accept(observationValueVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ObservationValue createObservationValue(Set<Statement> statements) {
		String id = null;

		for (Statement statement : statements) {
			URI p = statement.getPredicate();
			Value o = statement.getObject();

			if (p.equals(RDF.TYPE) && o.equals(SSN.asURI.ObservationValue)) {
				id = statement.getSubject().stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find observation value id; expected type SSN:ObservationValue [statements = "
						+ statements + "]");

			return null;
		}

		for (Statement statement : statements) {
			Resource s = statement.getSubject();
			URI p = statement.getPredicate();

			if (s.stringValue().equals(id)
					&& p.equals(DUL.asURI.hasRegionDataValue)) {
				Literal o = (Literal) statement.getObject();
				if (o.getDatatype().equals(XMLSchema.DOUBLE))
					return createObservationValueDouble(statements);
			}
		}

		if (log.isLoggable(Level.WARNING))
			log.warning("Could not create observation value; unrecognized datatype [statements = "
					+ statements + "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(ObservationValueDouble value) {
		Set<Statement> ret = new HashSet<Statement>();

		Double v = value.getValue();

		ret.add(_statement(_uri(value.getId()), RDF.TYPE, _uri(value.getType())));

		if (v == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No value [v = null]");
		} else {
			ret.add(_statement(_uri(value.getId()),
					DUL.asURI.hasRegionDataValue,
					_typedLiteral(v.toString(), XMLSchema.DOUBLE)));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ObservationValueDouble createObservationValueDouble(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			URI p = statement.getPredicate();
			Value o = statement.getObject();

			if (p.equals(RDF.TYPE) && o.equals(SSN.asURI.ObservationValue)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find observation value double id; expected type SSN:ObservationValue [statements = "
						+ statements + "]");

			return null;
		}

		for (Statement statement : statements) {
			Resource s = statement.getSubject();
			URI p = statement.getPredicate();

			if (s.stringValue().equals(id)
					&& p.equals(DUL.asURI.hasRegionDataValue)) {
				Literal o = (Literal) statement.getObject();
				if (o.getDatatype().equals(XMLSchema.DOUBLE))
					return new ObservationValueDouble(id, type, o.doubleValue());
			}
		}

		if (log.isLoggable(Level.WARNING))
			log.warning("Could not create observation value; unrecognized datatype [statements = "
					+ statements + "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(SensorObservation observation) {
		Set<Statement> ret = new HashSet<Statement>();

		Sensor sensor = observation.getSensor();
		Property property = observation.getProperty();
		Feature feature = observation.getFeature();
		Stimulus stimulus = observation.getStimulus();
		SensorOutput output = observation.getSensorOutput();
		TemporalLocation temporalLocation = observation.getTemporalLocation();
		SpatialLocation spatialLocation = observation.getSpatialLocation();

		observationId = observation.getId();

		ret.add(_statement(_uri(observationId), RDF.TYPE,
				_uri(observation.getType())));

		if (sensor == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No sensor [observation = " + observation + "]");
		} else {
			ret.add(_statement(_uri(observationId), SSN.asURI.observedBy,
					_uri(sensor.getId())));
			ret.addAll(createRepresentation(sensor));
		}

		if (property == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No property [observation = " + observation + "]");
		} else {
			ret.add(_statement(_uri(observationId), SSN.asURI.observedProperty,
					_uri(property.getId())));
			ret.addAll(createRepresentation(property));
		}

		if (feature == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No feature [observation = " + observation + "]");
		} else {
			ret.add(_statement(_uri(observationId),
					SSN.asURI.featureOfInterest, _uri(feature.getId())));
			ret.addAll(createRepresentation(feature));
		}

		if (stimulus != null) {
			ret.add(_statement(_uri(observationId), DUL.asURI.includesEvent,
					_uri(stimulus.getId())));
			ret.addAll(createRepresentation(stimulus));
		}

		if (output == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No output [observation = " + observation + "]");
		} else {
			ret.add(_statement(_uri(observationId),
					SSN.asURI.observationResult, _uri(output.getId())));
			ret.addAll(createRepresentation(output));
		}

		if (temporalLocation == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No temporal location [observation = "
						+ temporalLocation + "]");
		} else {
			ret.add(_statement(_uri(observationId),
					SSN.asURI.observationResultTime,
					_uri(temporalLocation.getId())));
			ret.addAll(entityRepresentationRdfTime
					.createRepresentation(temporalLocation));
		}

		if (spatialLocation != null) {
			statements = ret;
			spatialLocation.accept(spatialLocationVisitor);
			ret.addAll(entityRepresentationRdfGeo
					.createRepresentation(spatialLocation));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public SensorObservation createSensorObservation(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			URI p = statement.getPredicate();
			Value o = statement.getObject();

			if (p.equals(RDF.TYPE) && o.equals(WOE.asURI.SensorObservation)) {
				id = statement.getSubject().stringValue();
				type = o.stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find observation id; expected type WOE:SensorObservation [statements = "
						+ statements + "]");

			return null;
		}

		SensorObservation ret = new SensorObservationBase(id, type);

		ret.setSensor(createSensor(statements));
		ret.setProperty(createProperty(statements));
		ret.setFeature(createFeature(statements));
		ret.setSensorOutput(createSensorOutput(statements));
		ret.setTemporalLocation(entityRepresentationRdfTime
				.createTemporalLocation(statements));
		ret.setSpatialLocation(entityRepresentationRdfGeo
				.createSpatialLocation(statements));

		return ret;
	}

	private class ThisSpatialLocationVisitor implements SpatialLocationVisitor {

		@Override
		public void visit(SpatialLocationPlace location) {
			statements
					.add(_statement(_uri(observationId),
							WOE.asURI.observationResultLocation,
							_uri(location.getId())));
		}

		@Override
		public void visit(SpatialLocationRegion location) {
			statements
					.add(_statement(_uri(observationId),
							WOE.asURI.observationResultLocation,
							_uri(location.getId())));
		}

	}

	private class ThisObservationValueVisitor implements
			ObservationValueVisitor {

		@Override
		public void visit(ObservationValueDouble value) {
			statements.addAll(createRepresentation(value));
		}

	}

}
