/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.representation.rdf;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.joda.time.Period;
import org.joda.time.format.ISOPeriodFormat;
import org.joda.time.format.PeriodFormatter;
import org.openrdf.model.Literal;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.XMLSchema;

import fi.uef.envi.wavellite.entity.core.SpatialLocation;
import fi.uef.envi.wavellite.entity.core.TemporalLocation;
import fi.uef.envi.wavellite.entity.situation.Attribute;
import fi.uef.envi.wavellite.entity.situation.AttributeVisitor;
import fi.uef.envi.wavellite.entity.situation.RelevantObjectVisitor;
import fi.uef.envi.wavellite.entity.situation.Value;
import fi.uef.envi.wavellite.entity.situation.ElementaryInfon;
import fi.uef.envi.wavellite.entity.situation.Polarity;
import fi.uef.envi.wavellite.entity.situation.Relation;
import fi.uef.envi.wavellite.entity.situation.RelevantIndividual;
import fi.uef.envi.wavellite.entity.situation.RelevantObject;
import fi.uef.envi.wavellite.entity.situation.Situation;
import fi.uef.envi.wavellite.entity.situation.ValueVisitor;
import fi.uef.envi.wavellite.entity.situation.base.ElementaryInfonBase;
import fi.uef.envi.wavellite.entity.situation.base.RelationBase;
import fi.uef.envi.wavellite.entity.situation.base.RelevantIndividualBase;
import fi.uef.envi.wavellite.entity.situation.base.SituationBase;
import fi.uef.envi.wavellite.entity.situation.base.ValueDouble;
import fi.uef.envi.wavellite.entity.situation.base.ValueInteger;
import fi.uef.envi.wavellite.entity.situation.base.AttributeSpatialLocation;
import fi.uef.envi.wavellite.entity.situation.base.ValuePeriod;
import fi.uef.envi.wavellite.entity.situation.base.ValueString;
import fi.uef.envi.wavellite.entity.situation.base.AttributeTemporalLocation;
import fi.uef.envi.wavellite.entity.situation.base.AttributeUri;
import fi.uef.envi.wavellite.entity.situation.base.AttributeValue;
import fi.uef.envi.wavellite.entity.situation.base.ValueUri;
import fi.uef.envi.wavellite.vocabulary.STO;
import fi.uef.envi.wavellite.vocabulary.WOE;

/**
 * <p>
 * Title: EntityRepresentationRdfSto
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Representation
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class EntityRepresentationRdfSto extends EntityRepresentationRdf {

	private Set<Statement> statements;
	private Integer anchorN = 1;
	private String infonId = null;
	private String individualId = null;
	private String attributeId = null;
	private AttributeVisitor attributeVisitor;
	private ValueVisitor attributeValueVisitor;
	private RelevantObjectVisitor relevantObjectVisitor;
	private EntityRepresentationRdfGeo entityRepresentationRdfGeo;
	private EntityRepresentationRdfTime entityRepresentationRdfTime;

	private final static PeriodFormatter pf = ISOPeriodFormat.standard();
	private final static Logger log = Logger
			.getLogger(EntityRepresentationRdfSto.class.getName());

	public EntityRepresentationRdfSto() {
		this(null);
	}

	public EntityRepresentationRdfSto(String ns) {
		this(ns, null, null);
	}

	public EntityRepresentationRdfSto(String ns,
			EntityRepresentationRdfGeo entityRepresentationRdfGeo,
			EntityRepresentationRdfTime entityRepresentationRdfTime) {
		super(ns);

		if (entityRepresentationRdfGeo == null)
			this.entityRepresentationRdfGeo = new EntityRepresentationRdfGeo(ns);
		else
			this.entityRepresentationRdfGeo = entityRepresentationRdfGeo;

		if (entityRepresentationRdfTime == null)
			this.entityRepresentationRdfTime = new EntityRepresentationRdfTime(
					ns);
		else
			this.entityRepresentationRdfTime = entityRepresentationRdfTime;

		this.attributeVisitor = new ThisAttributeVisitor();
		this.attributeValueVisitor = new ThisAttributeValueVisitor();
		this.relevantObjectVisitor = new ThisRelevantObjectVisitor();
	}

	@Override
	public void setNamespace(String ns) {
		super.setNamespace(ns);

		entityRepresentationRdfGeo.setNamespace(ns);
		entityRepresentationRdfTime.setNamespace(ns);
	}

	@Override
	public Set<Statement> createRepresentation(SpatialLocation location) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.addAll(entityRepresentationRdfGeo.createRepresentation(location));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Set<Statement> createRepresentation(Situation situation) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(situation.getId()), RDF.TYPE,
				_uri(situation.getType())));

		for (ElementaryInfon infon : situation.getSupportedInfons()) {
			ret.add(_statement(_uri(situation.getId()),
					STO.asURI.supportedInfon, _uri(infon.getId())));
			ret.addAll(createRepresentation(infon));
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Situation createSituation(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.Situation)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = STO.Situation;
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Cannot create situation; multiple STO.Situation(s) found [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Cannot create situation; expected type STO.Situation [statements = "
						+ statements + "]");

			return null;
		}

		Situation ret = new SituationBase(id, type);

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.ElementaryInfon)) {
				Set<Statement> result = new HashSet<Statement>();
				getTransitiveStatementsMatchingSubject(statement.getSubject()
						.stringValue(), statements, result);
				ElementaryInfon infon = createElementaryInfon(result);

				if (infon == null) {
					if (log.isLoggable(Level.WARNING))
						log.warning("Returned elementary infon is null [statements = "
								+ statements + "]");
				} else {
					ret.addSupportedInfon(infon);
				}
			}
		}

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(ElementaryInfon infon) {
		Set<Statement> ret = new HashSet<Statement>();

		Relation relation = infon.getRelation();
		Polarity polarity = infon.getPolarity();

		anchorN = 1;
		infonId = infon.getId();

		ret.add(_statement(_uri(infonId), RDF.TYPE, _uri(infon.getType())));

		if (relation == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No relation [infon = " + infon + "]");
		} else {
			ret.add(_statement(_uri(infonId), STO.asURI.relation,
					_uri(relation.getId())));
			ret.addAll(createRepresentation(relation));
		}

		for (RelevantObject object : infon.getRelevantObjects()) {
			ret.addAll(createRepresentation(object));
		}

		if (polarity == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No polarity [polarity = " + polarity + "]");
		} else {
			if (polarity.equals(Polarity.True))
				ret.add(_statement(_uri(infonId), STO.asURI.polarity,
						_uri(STO._1)));
			else
				ret.add(_statement(_uri(infonId), STO.asURI.polarity,
						_uri(STO._0)));
		}

		anchorN = null;
		infonId = null;

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ElementaryInfon createElementaryInfon(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.ElementaryInfon)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = STO.ElementaryInfon;
				} else {
					log.severe("Cannot create elementary infon; multiple STO.ElementaryInfon(s) found [statements = "
							+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create elementary infon; expected type STO.ElementaryInfon [statements = "
						+ statements + "]");

			return null;
		}

		ElementaryInfon ret = new ElementaryInfonBase(id, type);

		ret.setRelation(createRelation(statements));
		ret.setPolarity(createPolarity(statements));

		Set<Statement> anchors = new HashSet<Statement>();

		for (Statement statement : statements) {
			String fragment = statement.getPredicate().getLocalName();

			if (!fragment.startsWith("anchor"))
				continue;

			anchors.add(statement);
		}

		// We need to sort the anchor statements by anchor1, anchor2, anchor3,
		// ... and add them to the elementary infon returned here in that order
		// to preserve the correct sequence of relevant objects in the infon
		Statement[] sorted = new Statement[anchors.size()];

		for (Statement statement : anchors) {
			String fragment = statement.getPredicate().getLocalName();

			if (!fragment.startsWith("anchor")) {
				if (log.isLoggable(Level.SEVERE))
					log.severe("Expected statement with anchorN predicate [statement = "
							+ statement + "]");
			} else {
				// e.g. anchor1
				sorted[Integer.valueOf(fragment.replace("anchor", "")) - 1] = statement;
			}
		}

		for (int i = 0; i < sorted.length; i++) {
			Statement statement = sorted[i];

			Set<Statement> result = new HashSet<Statement>();

			// This is important for the case in which the infon anchors a URI
			result.add(statement);

			getTransitiveStatementsMatchingSubject(statement.getObject()
					.stringValue(), statements, result);

			RelevantObject object = createRelevantObject(result);

			if (object == null) {
				if (log.isLoggable(Level.SEVERE))
					log.severe("Returned relevant object is null [result = "
							+ result + "]");
			} else {
				ret.addRelevantObject(object);
			}
		}

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(Relation relation) {
		Set<Statement> ret = new HashSet<Statement>();

		ret.add(_statement(_uri(relation.getId()), RDF.TYPE,
				_uri(relation.getType())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Relation createRelation(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.Relation)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = STO.Relation;
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Cannot create relation; multiple STO.Relation(s) found [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create relation; expected type STO.Relation [statements = "
						+ statements + "]");

			return null;
		}

		return new RelationBase(id, type);
	}

	@Override
	public Polarity createPolarity(Set<Statement> statements) {
		Polarity ret = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(STO.asURI.polarity)) {
				if (ret == null) {
					if (statement.getObject().equals(STO.asURI._0))
						ret = Polarity.False;
					else if (statement.getObject().equals(STO.asURI._1))
						ret = Polarity.True;
					else {
						if (log.isLoggable(Level.SEVERE))
							log.severe("Cannot create polarity; unrecognized [statements = "
									+ statements + "]");

						return null;
					}
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Cannot create polarity; multiple STO.polarity found [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(RelevantObject object) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		object.accept(relevantObjectVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public RelevantObject createRelevantObject(Set<Statement> statements) {
		// TODO: this is rather ugly here; unfortunately so far it seems
		// necessary to handle all possible relevant objects and because the
		// statements returned by the SPARQL construct query for situations does
		// not include statements about which relevant objects are attributes
		// (uri, temporal and spatial locations)

		// This is a hack, assuming it is a statement for an infon anchoring a
		// URI, hence it is an AttributeUri but it has no type information
		// (which is the case for the SPARQL construct query)
		if (statements.size() == 1) {
			Statement s = statements.toArray(new Statement[] {})[0];
			statements.add(f.createStatement((URI) s.getObject(), RDF.TYPE,
					STO.asURI.Attribute));
		}

		// Make sure temporal locations and spatial locations are marked as
		// attributes
		boolean isTimeInterval = false;

		Set<Statement> set = new HashSet<Statement>(statements);

		for (Statement statement : set) {
			if (!statement.getPredicate().equals(RDF.TYPE))
				continue;

			org.openrdf.model.Value o = statement.getObject();

			if (o.equals(WOE.asURI.TimeInterval)) {
				statements.add(f.createStatement(statement.getSubject(),
						RDF.TYPE, STO.asURI.Attribute));
				isTimeInterval = true;
			} else if (o.equals(WOE.asURI.SpatialPlace)
					|| o.equals(WOE.asURI.SpatialRegion)) {
				statements.add(f.createStatement(statement.getSubject(),
						RDF.TYPE, STO.asURI.Attribute));
			}
		}

		if (!isTimeInterval) {
			for (Statement statement : set) {
				if (!statement.getPredicate().equals(RDF.TYPE))
					continue;

				org.openrdf.model.Value o = statement.getObject();

				if (o.equals(WOE.asURI.TimePoint)) {
					statements.add(f.createStatement(statement.getSubject(),
							RDF.TYPE, STO.asURI.Attribute));
				}
			}
		}

		// This has to be processed first because relevant individuals may have
		// attributes (including temporal location etc.)
		for (Statement statement : statements) {
			URI p = statement.getPredicate();
			org.openrdf.model.Value o = statement.getObject();

			if (p.equals(RDF.TYPE) && o.equals(STO.asURI.RelevantIndividual)) {
				Set<Statement> result = new HashSet<Statement>();
				getTransitiveStatementsMatchingSubject(statement.getSubject()
						.stringValue(), statements, result);
				return createRelevantIndividual(result);
			}
		}

		// As we processed relevant individuals, these are now relevant objects
		// that are directly anchored by the infon, e.g. URIs and temporal and
		// spatial locations, or values
		for (Statement statement : statements) {
			if (!statement.getPredicate().equals(RDF.TYPE))
				continue;

			org.openrdf.model.Value o = statement.getObject();

			if (o.equals(STO.asURI.Attribute)) {
				Set<Statement> result = new HashSet<Statement>();
				getTransitiveStatementsMatchingSubject(statement.getSubject()
						.stringValue(), statements, result);
				return (RelevantObject) createAttribute(result);
			} else if (o.equals(STO.asURI.Value)) {
				Set<Statement> result = new HashSet<Statement>();
				getTransitiveStatementsMatchingSubject(statement.getSubject()
						.stringValue(), statements, result);
				return (RelevantObject) createValue(result);
			}
		}

		if (log.isLoggable(Level.SEVERE))
			log.severe("Cannot cerate relevant object [statements = "
					+ statements + "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(RelevantIndividual individual) {
		Set<Statement> ret = new HashSet<Statement>();

		individualId = individual.getId();

		if (infonId != null)
			ret.add(_statement(_uri(infonId), _uri(STO.ns + "#anchor"
					+ (anchorN++)), _uri(individualId)));

		ret.add(_statement(_uri(individualId), RDF.TYPE,
				_uri(individual.getType())));

		for (Attribute attribute : individual.getAttributes()) {
			ret.addAll(createRepresentation(attribute));
		}

		individualId = null;

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public RelevantIndividual createRelevantIndividual(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(
							STO.asURI.RelevantIndividual)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = STO.RelevantIndividual;
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Cannot create relevant individual; multiple STO.RelevantIndividual(s) found [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create relevant individual; expected type STO.RelevantIndividua [statements = "
						+ statements + "]");

			return null;
		}

		RelevantIndividual ret = new RelevantIndividualBase(id, type);

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.Attribute)) {
				Set<Statement> result = new HashSet<Statement>();
				getTransitiveStatementsMatchingSubject(statement.getSubject()
						.stringValue(), statements, result);
				Attribute attribute = createAttribute(result);

				if (attribute == null) {
					if (log.isLoggable(Level.WARNING))
						log.warning("Returned attribute is null [statements = "
								+ statements + "]");
				} else {
					ret.addAttribute(attribute);
				}
			}
		}

		return ret;
	}

	@Override
	public Set<Statement> createRepresentation(Attribute attribute) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		attribute.accept(attributeVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Attribute createAttribute(Set<Statement> statements) {
		String id = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.Attribute)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Cannote create attribute; multiple STO.Attribute(s) found [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create attribute; expected type STO.Attribute [statements = "
						+ statements + "]");

			return null;
		}

		Set<Statement> results = new HashSet<Statement>();
		getTransitiveStatementsMatchingSubject(id, statements, results);

		if (results.size() == 1)
			return createAttributeUri(results);

		for (Statement result : results) {
			if (result.getSubject().stringValue().equals(id)
					&& result.getPredicate()
							.equals(STO.asURI.hasAttributeValue)) {
				return createAttributeValue(results);
			}

			if (result.getPredicate().equals(RDF.TYPE)
					&& (result.getObject().equals(WOE.asURI.TimePoint) || result
							.getObject().equals(WOE.asURI.TimeInterval))) {
				return createAttributeTemporalLocation(results);
			}

			if (result.getPredicate().equals(RDF.TYPE)
					&& (result.getObject().equals(WOE.asURI.SpatialPlace) || result
							.getObject().equals(WOE.asURI.SpatialRegion))) {
				return createAttributeSpatialLocation(results);
			}

		}

		if (log.isLoggable(Level.SEVERE))
			log.severe("Cannote create attribute [statements = " + statements
					+ "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(AttributeSpatialLocation object) {
		Set<Statement> ret = new HashSet<Statement>();

		SpatialLocation location = object.getValue();

		if (location == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No attribute value [attribute = " + object + "]");

			return Collections.unmodifiableSet(ret);
		}

		ret.addAll(entityRepresentationRdfGeo.createRepresentation(location));

		// It is important to test this like this because if both individualId
		// and infonId are not null, which can happen in infons with with
		// multiple objects, then both statements will be added, which cannot be
		// (the spatial location is either a relevant object or an attribute of
		// a relevant individual)
		if (individualId != null) {
			// Also include the canonical hasAttribute
			ret.add(_statement(_uri(individualId), STO.asURI.hasAttribute,
					_uri(location.getId())));
			ret.add(_statement(_uri(individualId), _uri(object.getName()),
					_uri(location.getId())));

			return Collections.unmodifiableSet(ret);
		}

		if (infonId != null)
			ret.add(_statement(_uri(infonId), _uri(STO.ns + "#anchor"
					+ (anchorN++)), _uri(location.getId())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public AttributeSpatialLocation createAttributeSpatialLocation(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.Attribute)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = STO.Attribute;
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Cannot create attribute spatial location; multiple Attribute(s) found [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create attribute spatial location; expected type STO.Attribute [statements = "
						+ statements + "]");

			return null;
		}

		return new AttributeSpatialLocation(id, type,
				entityRepresentationRdfGeo.createSpatialLocation(statements));
	}

	@Override
	public Set<Statement> createRepresentation(AttributeTemporalLocation object) {
		Set<Statement> ret = new HashSet<Statement>();

		TemporalLocation location = object.getValue();

		if (location == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No attribute value [attribute = " + object + "]");

			return Collections.unmodifiableSet(ret);
		}

		ret.addAll(entityRepresentationRdfTime.createRepresentation(location));

		// It is important to test this like this because if both individualId
		// and infonId are not null, which can happen in infons with with
		// multiple objects, then both statements will be added, which cannot be
		// (the temporal location is either a relevant object or an attribute of
		// a relevant individual)
		if (individualId != null) {
			// Also include the canonical hasAttribute property
			ret.add(_statement(_uri(individualId), STO.asURI.hasAttribute,
					_uri(location.getId())));
			ret.add(_statement(_uri(individualId), _uri(object.getName()),
					_uri(location.getId())));

			return Collections.unmodifiableSet(ret);
		}

		if (infonId != null)
			ret.add(_statement(_uri(infonId), _uri(STO.ns + "#anchor"
					+ (anchorN++)), _uri(location.getId())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public AttributeTemporalLocation createAttributeTemporalLocation(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.Attribute)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = STO.Attribute;
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Cannot create attribute temporal location; multiple Attribute(s) found [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create attribute temporal location; expected type STO.Attribute [statements = "
						+ statements + "]");

			return null;
		}

		return new AttributeTemporalLocation(id, type,
				entityRepresentationRdfTime.createTemporalLocation(statements));
	}

	@Override
	public Set<Statement> createRepresentation(AttributeUri object) {
		Set<Statement> ret = new HashSet<Statement>();

		URI uri = object.getValue();

		if (uri == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No attribute value [attribute = " + object + "]");

			return Collections.unmodifiableSet(ret);
		}

		ret.add(_statement(uri, RDF.TYPE, _uri(object.getType())));

		// It is important to test this like this because if both individualId
		// and infonId are not null, which can happen in infons with with
		// multiple objects, then both statements will be added, which cannot be
		// (the URI is either a relevant object or an attribute of a relevant
		// individual)
		if (individualId != null) {
			// Also include the canonical hasAttribute property
			ret.add(_statement(_uri(individualId), STO.asURI.hasAttribute, uri));
			ret.add(_statement(_uri(individualId), _uri(object.getName()), uri));

			return Collections.unmodifiableSet(ret);
		}

		if (infonId != null)
			ret.add(_statement(_uri(infonId), _uri(STO.ns + "#anchor"
					+ (anchorN++)), uri));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public AttributeUri createAttributeUri(Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannote create attribute uri; expected singleton [statements = "
						+ statements + "]");

			return null;
		}

		Statement s = statements.toArray(new Statement[] {})[0];

		return new AttributeUri((URI) s.getSubject());
	}

	@Override
	public Set<Statement> createRepresentation(AttributeValue attribute) {
		Set<Statement> ret = new HashSet<Statement>();

		Value value = attribute.getValue();

		attributeId = attribute.getId();

		ret.add(_statement(_uri(attributeId), RDF.TYPE,
				_uri(attribute.getType())));

		if (individualId != null) {
			// Also include the canonical hasAttribute property
			ret.add(_statement(_uri(individualId), STO.asURI.hasAttribute, _uri(attributeId)));
			ret.add(_statement(_uri(individualId), _uri(attribute.getName()),
					_uri(attributeId)));
		}

		if (value == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No attribute value [attribute = " + attribute
						+ "]");

			return Collections.unmodifiableSet(ret);
		}

		ret.addAll(createRepresentation(value));

		ret.add(_statement(_uri(attributeId), STO.asURI.hasAttributeValue,
				_uri(value.getId())));

		attributeId = null;

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public AttributeValue createAttributeValue(Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			if (statement.getPredicate().equals(RDF.TYPE)
					&& statement.getObject().equals(STO.asURI.Attribute)) {
				if (id == null) {
					id = statement.getSubject().stringValue();
					type = STO.Attribute;
				} else {
					if (log.isLoggable(Level.SEVERE))
						log.severe("Cannot create attribute value; multiple STO.Attribute(s) found [statements = "
								+ statements + "]");

					return null;
				}
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create attribute value; expected type STO.Attribute [statements = "
						+ statements + "]");

			return null;
		}

		return new AttributeValue(id, type, createValue(statements));
	}

	@Override
	public Set<Statement> createRepresentation(Value value) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		value.accept(attributeValueVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public Value createValue(Set<Statement> statements) {
		for (Statement statement : statements) {
			if (!statement.getPredicate().equals(STO.asURI.attributeValue))
				continue;

			org.openrdf.model.Value o = statement.getObject();

			if (o instanceof Literal) {
				Literal l = (Literal) o;
				URI dt = l.getDatatype();

				if (dt == null) {
					if (log.isLoggable(Level.WARNING))
						log.warning("Datatype is null; assumed attribute value String [statement = "
								+ statements + "]");

					return createValueString(Collections.singleton(statement));
				}

				if (dt.equals(XMLSchema.DOUBLE))
					return createValueDouble(Collections.singleton(statement));

				if (dt.equals(XMLSchema.INTEGER) || dt.equals(XMLSchema.INT))
					return createValueInteger(Collections.singleton(statement));

				if (dt.equals(XMLSchema.STRING))
					return createValueString(Collections.singleton(statement));

				if (dt.equals(XMLSchema.DURATION))
					return createValuePeriod(Collections.singleton(statement));
			}
		}

		if (log.isLoggable(Level.SEVERE))
			log.severe("Cannot create value [statements = " + statements + "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(ValueDouble value) {
		Set<Statement> ret = new HashSet<Statement>();

		Double v = value.getValue();

		ret.add(_statement(_uri(value.getId()), RDF.TYPE, _uri(value.getType())));

		if (v == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No value [v = null]");

			return Collections.unmodifiableSet(ret);
		}

		ret.add(_statement(_uri(value.getId()), STO.asURI.attributeValue,
				_typedLiteral(Double.toString(v), XMLSchema.DOUBLE)));

		if (attributeId != null) {
			return Collections.unmodifiableSet(ret);
		}

		if (infonId != null)
			ret.add(_statement(_uri(infonId), _uri(STO.ns + "#anchor"
					+ (anchorN++)), _uri(value.getId())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ValueDouble createValueDouble(Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create value double; expected singleton [statements = "
						+ statements + "]");

			return null;
		}

		Statement s = statements.toArray(new Statement[] {})[0];

		return new ValueDouble(s.getSubject().stringValue(), Double.valueOf(s
				.getObject().stringValue()));
	}

	@Override
	public Set<Statement> createRepresentation(ValueString value) {
		Set<Statement> ret = new HashSet<Statement>();

		String v = value.getValue();

		ret.add(_statement(_uri(value.getId()), RDF.TYPE, _uri(value.getType())));

		if (v == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No value [v = null]");

			return Collections.unmodifiableSet(ret);
		}

		ret.add(_statement(_uri(value.getId()), STO.asURI.attributeValue,
				_typedLiteral(v, XMLSchema.STRING)));

		if (attributeId != null) {
			return Collections.unmodifiableSet(ret);
		}

		if (infonId != null)
			ret.add(_statement(_uri(infonId), _uri(STO.ns + "#anchor"
					+ (anchorN++)), _uri(value.getId())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ValueString createValueString(Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create value string; expected singleton [statements = "
						+ statements + "]");

			return null;
		}

		Statement s = statements.toArray(new Statement[] {})[0];

		return new ValueString(s.getSubject().stringValue(), s.getObject()
				.stringValue());
	}

	@Override
	public Set<Statement> createRepresentation(ValueInteger value) {
		Set<Statement> ret = new HashSet<Statement>();

		Integer v = value.getValue();

		ret.add(_statement(_uri(value.getId()), RDF.TYPE, _uri(value.getType())));

		if (v == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No value [v = null]");

			return Collections.unmodifiableSet(ret);
		}

		ret.add(_statement(_uri(value.getId()), STO.asURI.attributeValue,
				_typedLiteral(v.toString(), XMLSchema.INT)));

		if (attributeId != null) {
			return Collections.unmodifiableSet(ret);
		}

		if (infonId != null)
			ret.add(_statement(_uri(infonId), _uri(STO.ns + "#anchor"
					+ (anchorN++)), _uri(value.getId())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ValueInteger createValueInteger(Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create value integer; expected singleton [statements = "
						+ statements + "]");

			return null;
		}

		Statement s = statements.toArray(new Statement[] {})[0];

		return new ValueInteger(s.getSubject().stringValue(), Integer.valueOf(s
				.getObject().stringValue()));
	}

	@Override
	public Set<Statement> createRepresentation(ValuePeriod value) {
		Set<Statement> ret = new HashSet<Statement>();

		Period v = value.getValue();

		ret.add(_statement(_uri(value.getId()), RDF.TYPE, _uri(value.getType())));

		if (v == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No value [v = null]");

			return Collections.unmodifiableSet(ret);
		}

		ret.add(_statement(_uri(value.getId()), STO.asURI.attributeValue,
				_typedLiteral(pf.print(v), XMLSchema.DURATION)));

		if (attributeId != null) {
			return Collections.unmodifiableSet(ret);
		}

		if (infonId != null)
			ret.add(_statement(_uri(infonId), _uri(STO.ns + "#anchor"
					+ (anchorN++)), _uri(value.getId())));

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public ValuePeriod createValuePeriod(Set<Statement> statements) {
		if (statements.size() != 1) {
			if (log.isLoggable(Level.SEVERE))
				log.severe("Cannot create value period; expected singleton [statements = "
						+ statements + "]");

			return null;
		}

		Statement s = statements.toArray(new Statement[] {})[0];

		return new ValuePeriod(s.getSubject().stringValue(), pf.parsePeriod(s
				.getObject().stringValue()));
	}

	private class ThisAttributeVisitor implements AttributeVisitor {

		@Override
		public void visit(AttributeValue attribute) {
			statements.addAll(createRepresentation(attribute));
		}

		@Override
		public void visit(AttributeSpatialLocation attribute) {
			statements.addAll(createRepresentation(attribute));
		}

		@Override
		public void visit(AttributeTemporalLocation attribute) {
			statements.addAll(createRepresentation(attribute));
		}

		@Override
		public void visit(AttributeUri attribute) {
			statements.addAll(createRepresentation(attribute));
		}

	}

	private class ThisAttributeValueVisitor implements ValueVisitor {

		@Override
		public void visit(ValueDouble value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ValueString value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ValueUri value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ValueInteger value) {
			statements.addAll(createRepresentation(value));
		}

		@Override
		public void visit(ValuePeriod value) {
			statements.addAll(createRepresentation(value));
		}

	}

	private class ThisRelevantObjectVisitor implements RelevantObjectVisitor {

		@Override
		public void visit(RelevantIndividual object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(AttributeSpatialLocation object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(AttributeTemporalLocation object) {
			statements.addAll(createRepresentation(object));
		}
		
		@Override
		public void visit(AttributeUri object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValueDouble object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValueInteger object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValuePeriod object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValueString object) {
			statements.addAll(createRepresentation(object));
		}

		@Override
		public void visit(ValueUri object) {
			statements.addAll(createRepresentation(object));
		}

	}

}
