/*
 * Copyright (C) 2013 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.representation.rdf;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.joda.time.DateTime;
import org.joda.time.format.DateTimeFormatter;
import org.joda.time.format.ISODateTimeFormat;
import org.openrdf.model.Statement;
import org.openrdf.model.URI;
import org.openrdf.model.Value;
import org.openrdf.model.vocabulary.RDF;
import org.openrdf.model.vocabulary.XMLSchema;

import fi.uef.envi.wavellite.entity.core.TemporalLocation;
import fi.uef.envi.wavellite.entity.core.TemporalLocationVisitor;
import fi.uef.envi.wavellite.entity.core.base.TemporalLocationDateTime;
import fi.uef.envi.wavellite.entity.core.base.TemporalLocationInterval;
import fi.uef.envi.wavellite.vocabulary.Time;
import fi.uef.envi.wavellite.vocabulary.WOE;

/**
 * <p>
 * Title: EntityRepresentationRdfTime
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Representation
 * </p>
 * <p>
 * Copyright: Copyright (C) 2013
 * </p>
 * 
 * @author Markus Stocker
 */

public class EntityRepresentationRdfTime extends EntityRepresentationRdf {

	private Set<Statement> statements;
	private TemporalLocationVisitor temporalLocationVisitor;
	private final static Logger log = Logger
			.getLogger(EntityRepresentationRdfTime.class.getName());
	private static final DateTimeFormatter dtf1 = ISODateTimeFormat.dateTime()
			.withOffsetParsed();
	private static final DateTimeFormatter dtf2 = ISODateTimeFormat
			.dateTimeParser().withOffsetParsed();

	public EntityRepresentationRdfTime() {
		this(null);
	}

	public EntityRepresentationRdfTime(String ns) {
		super(ns);

		this.temporalLocationVisitor = new ThisTemporalLocationVisitor();
	}

	@Override
	public Set<Statement> createRepresentation(TemporalLocation location) {
		Set<Statement> ret = new HashSet<Statement>();

		statements = ret;
		location.accept(temporalLocationVisitor);

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public TemporalLocation createTemporalLocation(Set<Statement> statements) {
		// Search first for the existence of TimeInterval, because if so then
		// the set of statements also contains TimePoint(s). If you scan them
		// together you might miss the TimeInterval if in the set the TimePoint
		// comes first!
		for (Statement statement : statements) {
			if (!statement.getPredicate().equals(RDF.TYPE))
				continue;

			if (statement.getObject().equals(WOE.asURI.TimeInterval))
				return createTemporalLocationInterval(statements);
		}

		for (Statement statement : statements) {
			if (!statement.getPredicate().equals(RDF.TYPE))
				continue;

			if (statement.getObject().equals(WOE.asURI.TimePoint))
				return createTemporalLocationDateTime(statements);
		}

		if (log.isLoggable(Level.WARNING))
			log.warning("Cannot create temporal location; unrecognized type [statements = "
					+ statements + "]");

		return null;
	}

	@Override
	public Set<Statement> createRepresentation(TemporalLocationDateTime location) {
		Set<Statement> ret = new HashSet<Statement>();

		String locationId = location.getId();
		DateTime value = location.getValue();

		if (value == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No value [time = null");
		} else {
			ret.add(_statement(_uri(locationId), RDF.TYPE,
					_uri(location.getType())));
			ret.add(_statement(_uri(locationId), Time.asURI.inXSDDateTime,
					_typedLiteral(dtf1.print(value), XMLSchema.DATETIME)));

			// TODO model the relation inDateTime with DateTimeDescription
			// Note that the date time description has to be held by the
			// TemporalLocationDateTime object, because otherwise you get new
			// identifiers each time you represent the same location, which you
			// probably want to avoid in order to reduce the amount of triples
			// that represent the same thing.
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public TemporalLocationDateTime createTemporalLocationDateTime(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			Value o = statement.getObject();

			if (statement.getPredicate().equals(RDF.TYPE)
					&& o.equals(WOE.asURI.TimePoint)) {
				id = statement.getSubject().stringValue();
				type = statement.getObject().stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find temporal location date time id; expected type WOE:TimePoint [statements = "
						+ statements + "]");
		}

		String dateTime = null;

		for (Statement statement : statements) {
			if (statement.getSubject().stringValue().equals(id)
					&& statement.getPredicate()
							.equals(Time.asURI.inXSDDateTime))
				dateTime = statement.getObject().stringValue();
		}

		if (dateTime == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find XSD date time for id [id = " + id
						+ "; statements = " + statements + "]");

			return null;
		}

		// This can fail in Stardog because it drops the milliseconds. To avoid
		// this, create a database.properties file with index.literals.canonical
		// = false and create the database with the command: stardog-admin db
		// create -c database.properties -n test
		return new TemporalLocationDateTime(id, type,
				dtf2.parseDateTime(dateTime));
	}

	@Override
	public Set<Statement> createRepresentation(TemporalLocationInterval location) {
		Set<Statement> ret = new HashSet<Statement>();

		TemporalLocationDateTime start = location.getStart();
		TemporalLocationDateTime end = location.getEnd();

		if (start == null || end == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("No value [start = " + start + "; end = " + end
						+ "; location = " + location + "]");
		} else {
			String locationId = location.getId();

			ret.add(_statement(_uri(locationId), RDF.TYPE,
					_uri(location.getType())));
			ret.add(_statement(_uri(locationId), Time.asURI.hasBeginning,
					_uri(start.getId())));
			ret.add(_statement(_uri(locationId), Time.asURI.hasEnd,
					_uri(end.getId())));

			ret.addAll(createRepresentation(start));
			ret.addAll(createRepresentation(end));

			// TODO model the relation hasDurationDescription with
			// DurationDescription. Note that the date time description has to
			// be held by the TemporalLocationDateTime object, because otherwise
			// you get new identifiers each time you represent the same
			// location, which you probably want to avoid in order to reduce the
			// amount of triples that represent the same thing.
		}

		return Collections.unmodifiableSet(ret);
	}

	@Override
	public TemporalLocationInterval createTemporalLocationInterval(
			Set<Statement> statements) {
		String id = null;
		String type = null;

		for (Statement statement : statements) {
			Value o = statement.getObject();

			if (statement.getPredicate().equals(RDF.TYPE)
					&& o.equals(WOE.asURI.TimeInterval)) {
				id = statement.getSubject().stringValue();
				type = statement.getObject().stringValue();
			}
		}

		if (id == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find temporal location interval id; expected type WOE:TimeInterval [statements = "
						+ statements + "]");
		}

		String hasBeginningId = null;
		String hasEndId = null;

		for (Statement statement : statements) {
			if (!statement.getSubject().stringValue().equals(id))
				continue;

			URI p = statement.getPredicate();

			if (p.equals(Time.asURI.hasBeginning))
				hasBeginningId = statement.getObject().stringValue();
			else if (p.equals(Time.asURI.hasEnd))
				hasEndId = statement.getObject().stringValue();
		}

		if (hasBeginningId == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find interval beginning date time [id = "
						+ id + "; statements = " + statements + "]");

			return null;
		}

		if (hasEndId == null) {
			if (log.isLoggable(Level.WARNING))
				log.warning("Could not find interval end date time [id = " + id
						+ "; statements = " + statements + "]");

			return null;
		}

		Set<Statement> result1 = new HashSet<Statement>();
		Set<Statement> result2 = new HashSet<Statement>();

		getStatementsMatchingSubject(hasBeginningId, statements, result1);
		getStatementsMatchingSubject(hasEndId, statements, result2);

		return new TemporalLocationInterval(id, type,
				createTemporalLocationDateTime(result1),
				createTemporalLocationDateTime(result2));
	}

	private class ThisTemporalLocationVisitor implements
			TemporalLocationVisitor {

		@Override
		public void visit(TemporalLocationDateTime location) {
			statements.addAll(createRepresentation(location));
		}

		@Override
		public void visit(TemporalLocationInterval location) {
			statements.addAll(createRepresentation(location));
		}

	}

}
