/*
 * Copyright (C) 2014 see CREDITS.txt
 * All rights reserved.
 */

package fi.uef.envi.wavellite.server;

import java.awt.Color;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.text.SimpleDateFormat;
import java.util.Iterator;

import javax.servlet.ServletContext;
import javax.ws.rs.GET;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.WebApplicationException;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.StreamingOutput;
import javax.ws.rs.core.Variant;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartUtilities;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.DateAxis;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.renderer.xy.XYLineAndShapeRenderer;
import org.jfree.data.time.Minute;
import org.jfree.data.time.Month;
import org.jfree.data.time.TimeSeries;
import org.jfree.data.time.TimeSeriesCollection;
import org.jfree.ui.RectangleInsets;
import org.joda.time.DateTime;
import org.joda.time.format.ISODateTimeFormat;

import fi.uef.envi.wavellite.entity.observation.SensorObservation;
import fi.uef.envi.wavellite.server.handler.ParamHandler;
import fi.uef.envi.wavellite.server.handler.RequestHandler;
import fi.uef.envi.wavellite.server.handler.ResponseHandler;

/**
 * <p>
 * Title: ResourceObservation
 * </p>
 * <p>
 * Description:
 * </p>
 * <p>
 * Project: Wavellite Server
 * </p>
 * <p>
 * Copyright: Copyright (C) 2014
 * </p>
 * 
 * @author Markus Stocker
 */

@Path("{db}/sensor/observation")
public class ResourceObservation {

	@PathParam("db")
	private String db;

	@Context
	private ServletContext servletContext;

	@GET
	@Path("list")
	@Produces(MediaType.TEXT_PLAIN)
	public Response getListAsTextPlain(@QueryParam("sensorId") String sensorId,
			@QueryParam("propertyId") String propertyId,
			@QueryParam("featureId") String featureId,
			@QueryParam("from") String from, @QueryParam("to") String to) {
		Iterator<SensorObservation> it = RequestHandler.getSensorObservations(
				servletContext, db, ParamHandler.asSensor(sensorId),
				ParamHandler.asProperty(propertyId),
				ParamHandler.asFeature(featureId),
				ParamHandler.asInterval(from, to));

		StringBuffer sb = new StringBuffer();

		while (it.hasNext()) {
			SensorObservation observation = it.next();
			DateTime dateTime = observation.getTemporalLocation()
					.getValueAsDateTime();
			Double value = observation.getSensorOutput().getObservationValue()
					.getValueAsDouble();

			sb.append(ISODateTimeFormat.dateTime().withOffsetParsed()
					.print(dateTime)
					+ "  " + value + "\n");
		}

		return ResponseHandler.getTextResponse(sb.toString());
	}

	@GET
	@Path("plot")
	@Produces("image/png")
	public Response getListAsImagePng(@QueryParam("sensorId") String sensorId,
			@QueryParam("propertyId") String propertyId,
			@QueryParam("featureId") String featureId,
			@QueryParam("from") String from, @QueryParam("to") String to,
			@QueryParam("propertyLabel") String propertyLabel,
			@QueryParam("featureLabel") String featureLabel,
			@QueryParam("ylabel") String ylabel) {
		Iterator<SensorObservation> it = RequestHandler.getSensorObservations(
				servletContext, db, ParamHandler.asSensor(sensorId),
				ParamHandler.asProperty(propertyId),
				ParamHandler.asFeature(featureId),
				ParamHandler.asInterval(from, to));

		StringBuffer timeSeriesLabel = new StringBuffer();

		if (featureLabel == null)
			timeSeriesLabel.append(featureId);
		else
			timeSeriesLabel.append(featureLabel);

		timeSeriesLabel.append(" ");

		if (propertyLabel == null)
			timeSeriesLabel.append(propertyId);
		else
			timeSeriesLabel.append(propertyLabel);

		TimeSeries s1 = new TimeSeries(timeSeriesLabel.toString());

		while (it.hasNext()) {
			SensorObservation observation = it.next();
			DateTime time = observation.getTemporalLocation()
					.getValueAsDateTime();
			Double value = observation.getSensorOutput().getObservationValue()
					.getValueAsDouble();
			s1.add(new Minute(time.toDate()), value);
		}

		TimeSeriesCollection dataset = new TimeSeriesCollection();
		dataset.addSeries(s1);

		final JFreeChart chart = ChartFactory.createTimeSeriesChart("", // title
				"Time", // x-axis label
				ylabel, // y-axis label
				dataset, // data
				true, // create legend?
				true, // generate tooltips?
				false // generate URLs?
				);

		chart.setBackgroundPaint(Color.white);

		XYPlot plot = (XYPlot) chart.getPlot();
		plot.setBackgroundPaint(Color.lightGray);
		plot.setDomainGridlinePaint(Color.white);
		plot.setRangeGridlinePaint(Color.white);
		plot.setAxisOffset(new RectangleInsets(5.0, 5.0, 5.0, 5.0));
		plot.setDomainCrosshairVisible(true);
		plot.setRangeCrosshairVisible(true);

		XYItemRenderer r = plot.getRenderer();

		if (r instanceof XYLineAndShapeRenderer) {
			XYLineAndShapeRenderer renderer = (XYLineAndShapeRenderer) r;
			renderer.setBaseShapesVisible(true);
			renderer.setBaseShapesFilled(true);
			renderer.setDrawSeriesLineAsPath(true);
		}

		DateAxis axis = (DateAxis) plot.getDomainAxis();
		axis.setDateFormatOverride(new SimpleDateFormat("MMM dd HH:mm"));

		StreamingOutput stream = new StreamingOutput() {
			@Override
			public void write(OutputStream os) throws IOException,
					WebApplicationException {
				try {
					ChartUtilities.writeChartAsPNG(os, chart, 800, 600);
				} catch (IOException e) {
					throw new WebApplicationException(e);
				}
			}
		};

		return Response.ok(stream).build();
	}

}
